import uuid
from typing import List, Callable, Optional, Dict, Any

from .fitness import Fitness
from .primitive_node import PrimitiveNode
from .terminal import Terminal


class Individual:
    """ Collection of PrimitiveNodes which together specify a machine learning pipeline.

    Parameters
    ----------
    main_node: PrimitiveNode
        The first node of the individual (the estimator node).
    to_pipeline: Callable, optional (default=None)
         A function which can convert this individual into a machine learning pipeline.
         If not provided, the `pipeline` property will be unavailable.
    """

    def __init__(
        self, main_node: PrimitiveNode, to_pipeline: Optional[Callable] = None
    ):
        self.fitness: Optional[Fitness] = None
        self.main_node = main_node
        self.meta: Dict[str, Any] = dict()
        self._id = uuid.uuid4()
        self._to_pipeline = to_pipeline

    def __eq__(self, other):
        return isinstance(other, Individual) and other._id == self._id

    def __hash__(self):
        return hash(self._id)

    def __str__(self):
        return (
            f"Individual {self._id}\n"
            f"Pipeline: {self.pipeline_str()}\nFitness: {self.fitness}"
        )

    @property
    def pipeline(self):
        """ Calls the `to_pipeline` method on itself."""
        if self._to_pipeline is None:
            raise AttributeError(
                "pipeline not available because `to_pipeline` was not set on __init__."
            )
        return self._to_pipeline(self)

    def short_name(self, step_separator: str = ">"):
        """ str: e.g. "Binarizer>BernoulliNB" """
        return step_separator.join(
            [str(primitive._primitive) for primitive in reversed(self.primitives)]
        )

    def pipeline_str(self):
        """ str: e.g. "BernoulliNB(Binarizer(data, Binarizer.threshold=0.6), BernoulliNB.alpha=1.0)" """  # noqa: E501
        return str(self.main_node)

    @property
    def primitives(self) -> List[PrimitiveNode]:
        """ Lists all primitive nodes, starting with the Individual's main node. """
        primitives = [self.main_node]
        current_node = self.main_node._data_node
        while isinstance(current_node, PrimitiveNode):  # i.e. not DATA_TERMINAL
            primitives.append(current_node)
            current_node = current_node._data_node
        return primitives

    @property
    def terminals(self) -> List[Terminal]:
        """ Lists all terminals connected to the Individual's primitive nodes. """
        return [terminal for prim in self.primitives for terminal in prim._terminals]

    def replace_terminal(self, position: int, new_terminal: Terminal):
        """ Replace the terminal at `position` by `new_terminal` in-place.

        Parameters
        ----------
        position: int
            Position (in self.terminals) of the terminal to be replaced.
        new_terminal: Terminal
            The new terminal to replace the old one with. Must share output type.
        """
        scan_position = 0
        for primitive in self.primitives:
            if scan_position + len(primitive._terminals) > position:
                terminal_to_be_replaced = primitive._terminals[position - scan_position]
                if terminal_to_be_replaced.identifier == new_terminal.identifier:
                    primitive._terminals[position - scan_position] = new_terminal
                    return
                else:
                    raise ValueError(
                        f"New terminal does not share output type with the old."
                        f"Old: {terminal_to_be_replaced.identifier}"
                        f"New: {new_terminal.identifier}."
                    )
            else:
                scan_position += len(primitive._terminals)
        if scan_position < position:
            raise ValueError(
                f"Position {position} is out of range with {scan_position} terminals."
            )

    def replace_primitive(self, position: int, new_primitive: PrimitiveNode):
        """ Replace the PrimitiveNode at `position` by `new_primitive`.

        Parameters
        ----------
        position: int
            Position (in self.primitives) of the PrimitiveNode to be replaced.
        new_primitive: PrimitiveNode
            The new PrimitiveNode to replace the old one with. Must share output type.
        """
        last_primitive = None
        for i, primitive_node in enumerate(self.primitives):
            if i == position:
                if primitive_node._primitive.output != new_primitive._primitive.output:
                    raise ValueError("New primitive output type differs from old.")
                if isinstance(primitive_node._data_node, str):
                    new_primitive._data_node = primitive_node._data_node
                else:
                    new_primitive._data_node = primitive_node._data_node.copy()
                break
            else:
                last_primitive = primitive_node

        if last_primitive is None:
            self.main_node = new_primitive
        else:
            last_primitive._data_node = new_primitive

    def copy_as_new(self):
        """ Make deep copy of self, but with fitness None and assigned a new id. """
        return Individual(self.main_node.copy(), to_pipeline=self._to_pipeline)

    @classmethod
    def from_string(
        cls, string: str, primitive_set: dict, to_pipeline: Optional[Callable] = None
    ):
        """ Construct an Individual from its `pipeline_str` representation.

        Parameters
        ----------
        string: str
            String formatted as `Individual.pipeline_str`.
        primitive_set: dict
            The dictionary defining all Terminals and Primitives.
        to_pipeline: Callable, optional (default=None)
            The function to convert the Individual into a pipeline representation.
            If `None`, the individuals `pipeline` property will not be available.

        Returns
        -------
        Individual
            An individual as defined by `str`.
        """
        expression = PrimitiveNode.from_string(string, primitive_set)
        return cls(expression, to_pipeline=to_pipeline)



"""
Contains mutation functions for genetic programming.
Each mutation function takes an individual and modifies it in-place.
"""
import random
from functools import partial
from typing import Callable, Optional, cast, List, Dict

from gama.genetic_programming.components import PrimitiveNode
from .components import Individual, DATA_TERMINAL
from .operations import random_primitive_node


def mut_replace_terminal(individual: Individual, primitive_set: dict) -> None:
    """ Mutates an Individual in-place by replacing one of its Terminals.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict
    """

    def terminal_replaceable(index_terminal):
        _, terminal = index_terminal
        return len(primitive_set[terminal.identifier]) > 1

    terminals = list(filter(terminal_replaceable, enumerate(individual.terminals)))
    if len(terminals) == 0:
        raise ValueError("Individual has no terminals suitable for mutation.")

    terminal_index, old = random.choice(terminals)
    candidates = filter(lambda t: t.value != old.value, primitive_set[old.identifier])

    new_terminal = random.choice(list(candidates))
    individual.replace_terminal(terminal_index, new_terminal)


def mut_replace_primitive(individual: Individual, primitive_set: dict) -> None:
    """ Mutates an Individual in-place by replacing one of its Primitives.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict
    """

    def primitive_replaceable(index_primitive):
        _, primitive = index_primitive
        return len(primitive_set[primitive._primitive.output]) > 1

    primitives = list(filter(primitive_replaceable, enumerate(individual.primitives)))
    if primitives == 0:
        raise ValueError("Individual has no primitives suitable for replacement.")

    primitive_index, old_primitive_node = random.choice(primitives)
    primitive_node = random_primitive_node(
        output_type=old_primitive_node._primitive.output,
        primitive_set=primitive_set,
        exclude=old_primitive_node._primitive,
    )
    individual.replace_primitive(primitive_index, primitive_node)


def mut_shrink(
    individual: Individual, primitive_set: dict = None, shrink_by: Optional[int] = None
) -> None:
    """ Mutates an Individual in-place by removing any number of primitive nodes.

    Primitive nodes are removed from the preprocessing end.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict, optional
        Not used. Present to create a matching function signature with other mutations.
    shrink_by: int, optional (default=None)
        Number of primitives to remove.
        Must be at least one greater than the number of primitives in `individual`.
        If None, a random number of primitives is removed.
    """
    n_primitives = len(list(individual.primitives))
    if shrink_by is not None and n_primitives <= shrink_by:
        raise ValueError(f"Can't shrink size {n_primitives} individual by {shrink_by}.")
    if shrink_by is None:
        shrink_by = random.randint(1, n_primitives - 1)

    current_primitive_node = individual.main_node
    primitives_left = n_primitives - 1
    while primitives_left > shrink_by:
        current_primitive_node = cast(PrimitiveNode, current_primitive_node._data_node)
        primitives_left -= 1
    current_primitive_node._data_node = DATA_TERMINAL


def mut_insert(individual: Individual, primitive_set: dict) -> None:
    """ Mutate an Individual in-place by inserting a PrimitiveNode at a random location.

    The new PrimitiveNode will not be inserted as root node.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict
    """
    parent_node = random.choice(list(individual.primitives))
    new_primitive_node = random_primitive_node(
        output_type=DATA_TERMINAL, primitive_set=primitive_set
    )
    new_primitive_node._data_node = parent_node._data_node
    parent_node._data_node = new_primitive_node


def random_valid_mutation_in_place(
    individual: Individual, primitive_set: dict, max_length: Optional[int] = None
) -> Callable:
    """ Apply a random valid mutation in place.

    The random mutation can be one of:

     - mut_random_primitive
     - mut_random_terminal, if the individual has at least one
     - mutShrink, if individual has at least two primitives
     - mutInsert, if it would not exceed `new_max_length` when specified.

    Parameters
    ----------
    individual: Individual
      An individual to be mutated *in-place*.
    primitive_set: dict
      A dictionary defining the set of primitives and terminals.
    max_length: int, optional (default=None)
     If specified, impose a maximum length on the new individual.


    Returns
    -------
    Callable
        The mutation function used.
    """
    n_primitives = len(list(individual.primitives))
    available_mutations: List[Callable[[Individual, Dict], None]] = []
    if max_length is not None and n_primitives > max_length:
        available_mutations.append(
            partial(mut_shrink, shrink_by=n_primitives - max_length)
        )
    else:
        available_mutations.append(mut_replace_primitive)
        if max_length is None or n_primitives < max_length:
            available_mutations.append(mut_insert)
        if n_primitives > 1:
            available_mutations.append(mut_shrink)

        replaceable_terminals = filter(
            lambda t: len(primitive_set[t.identifier]) > 1, individual.terminals
        )
        if len(list(replaceable_terminals)) > 1:
            available_mutations.append(mut_replace_terminal)

    mut_fn = random.choice(available_mutations)
    mut_fn(individual, primitive_set)

    return mut_fn



import os
import time

import dash_table
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate
from plotly import graph_objects as go
from dash.dependencies import Input, Output, State

from gama.dashboard.components.cli_window import CLIWindow
from gama.dashboard.pages.base_page import BasePage
from gama.logging.GamaReport import GamaReport


class RunningPage(BasePage):
    def __init__(self):
        super().__init__(name="Running", alignment=1, starts_hidden=True)
        self.cli = None
        self.id = "running-page"
        self.report = None
        self.log_file = None

    def build_page(self, app, controller):
        self.cli = CLIWindow("cli", app)
        plot_area = self.plot_area()
        pl_viz = self.pipeline_viz()
        pl_list = self.pipeline_list()
        ticker = dcc.Interval(id="ticker", interval=5000)
        self._content = html.Div(
            id=self.id,
            children=[
                dbc.Row([dbc.Col(plot_area, width=8), dbc.Col(self.cli.html)]),
                dbc.Row([dbc.Col(pl_viz, width=4), dbc.Col(pl_list)]),
                ticker,
            ],
        )

        app.callback(
            [
                Output("evaluation-graph", "figure"),
                Output("pipeline-table", "data"),
                Output("pl-viz", "children"),
                Output("pipeline-table", "selected_rows"),
                Output("pipeline-table", "selected_row_ids"),
                Output("evaluation-graph", "clickData"),
            ],
            [Input("ticker", "n_intervals"), Input("running-page-store", "data")],
        )(self.update_page)

        app.callback(
            [Output("running-page-store", "data")],
            [
                Input("evaluation-graph", "clickData"),
                Input("pipeline-table", "selected_row_ids"),
            ],
            [State("running-page-store", "data")],
        )(self.update_selection)

        return self._content

    def update_selection(self, click_data, selected_row_ids, page_store):
        cell_selected = None if selected_row_ids is None else selected_row_ids[0]
        if click_data is None:
            click_selected = None
        else:
            click_selected = click_data["points"][0]["customdata"]
        selected = click_selected if click_selected is not None else cell_selected

        # Selected row ids and click data are always set back to None.
        # The value that is not None is the new value.
        if selected is not None:
            self.need_update = True
            page_store["selected_pipeline"] = selected
            return [page_store]
        # First call or sync call.
        raise PreventUpdate

    def update_page(self, _, page_store):
        if self.report is None:
            if self.log_file is None or not os.path.exists(self.log_file):
                raise PreventUpdate  # report does not exist
            else:
                self.report = GamaReport(self.log_file)
                if self.report.evaluations.empty:
                    raise PreventUpdate
        elif not self.report.update() and not self.need_update:
            raise PreventUpdate  # report is not updated

        start_update = time.time()
        selected_pipeline = page_store.get("selected_pipeline", None)
        evaluations = self.report.successful_evaluations

        self.need_update = False
        scatters = self.scatter_plot(
            evaluations, self.report.metrics, selected_pipeline
        )
        metric_one, metric_two = self.report.metrics
        metric_one_text = metric_one.replace("_", " ")

        figure = {
            "data": scatters,
            "layout": dict(
                hovermode="closest",
                clickmode="event+select",
                title=f"{metric_one_text} vs {metric_two}",
                xaxis=dict(title=metric_one_text),
                yaxis=dict(
                    title=metric_two, tickformat=",d"
                ),  # tickformat forces integer ticks for length,
                uirevision="never_reset_zoom",
            ),
        }

        pl_table_data = [
            {
                "pl": self.report.individuals[id_].short_name(" > "),
                "id": id_,
                "score": score,
            }
            for id_, score in zip(evaluations.id, evaluations[metric_one])
        ]
        row_id = [i for i, id_ in enumerate(evaluations.id) if id_ == selected_pipeline]

        def format_pipeline(ind):
            pipeline_elements = []
            for primitive_node in reversed(ind.primitives):
                pipeline_elements.append(html.B(str(primitive_node._primitive)))
                pipeline_elements.append(html.Br())
                for terminal in primitive_node._terminals:
                    pipeline_elements.append(f"    {terminal}")
                    pipeline_elements.append(html.Br())
            return pipeline_elements

        if selected_pipeline is None:
            pl_viz_data = None
        else:
            pl_viz_data = format_pipeline(self.report.individuals[selected_pipeline])

        print("Update complete in ", time.time() - start_update)
        return figure, pl_table_data, pl_viz_data, row_id, None, None

    def scatter_plot(self, evaluations, metrics, selected_pipeline: str = None):
        metric_one, metric_two = metrics

        # Marker size indicates recency of the evaluations,
        # recent evaluations are bigger.
        biggest_size = 25
        smallest_size = 5
        selected_size = 30
        d_size_min_max = biggest_size - smallest_size

        sizes = list(range(smallest_size, biggest_size))[-len(evaluations) :]
        if len(evaluations) > d_size_min_max:
            sizes = [smallest_size] * (len(evaluations) - d_size_min_max) + sizes
        if selected_pipeline is not None:
            sizes = [
                size if id_ != selected_pipeline else selected_size
                for size, id_ in zip(sizes, evaluations.id)
            ]

        default_color = "#301cc9"
        selected_color = "#c81818"

        colors = [
            default_color if id_ != selected_pipeline else selected_color
            for id_ in evaluations.id
        ]

        print(evaluations.head())
        print(evaluations[metric_one])
        print(evaluations[metric_two])
        all_scatter = go.Scatter(
            x=evaluations[metric_one],
            y=-evaluations[metric_two],
            mode="markers",
            marker={"color": colors, "size": sizes},
            name="all evaluations",
            text=[self.report.individuals[id_].short_name() for id_ in evaluations.id],
            customdata=evaluations.id,
        )
        return [all_scatter]

    def gama_started(self, process, log_file):
        self.cli.monitor(process)
        self.log_file = os.path.expanduser(log_file)

    def plot_area(self):
        scatter = dcc.Graph(
            id="evaluation-graph",
            figure={
                "data": [],
                "layout": dict(hovermode="closest", transition={"duration": 500},),
            },
        )
        return html.Div(
            scatter,
            style={
                "height": "100%",
                "box-shadow": "1px 1px 1px black",
                "padding": "2%",
            },
        )

    def pipeline_list(self):
        ta = dash_table.DataTable(
            id="pipeline-table",
            columns=[
                {"name": "Pipeline", "id": "pl"},
                {"name": "Score", "id": "score"},
            ],
            data=[],
            style_table={"maxHeight": "300px", "overflowY": "scroll"},
            filter_action="native",
            sort_action="native",
            row_selectable="single",
            persistence_type="session",
            persistence=True,
        )

        return html.Div(
            ta,
            style={
                "height": "100%",
                "box-shadow": "1px 1px 1px black",
                "padding": "2%",
            },
        )

    def pipeline_viz(self):
        return html.Div(
            id="pl-viz",
            style={
                "height": "100%",
                "box-shadow": "1px 1px 1px black",
                "padding": "2%",
                "whiteSpace": "pre-wrap",
            },
        )



import os
from datetime import datetime, timedelta
from typing import List, Tuple, Dict

import pandas as pd

from gama.configuration.classification import clf_config
from gama.configuration.parser import pset_from_config, merge_configurations
from gama.configuration.regression import reg_config
from gama.genetic_programming.components import Individual


pset, _ = pset_from_config(merge_configurations(clf_config, reg_config))


class GamaReport:
    """ Contains information parsed from a search captured by a GAMA analysis log. """

    def __init__(self, log_directory: str):
        """ Parse the logfile or log lines provided.

        Parameters
        ----------
        log_directory: str
            The directory with logs:
                - gama.log
                - evaluations.log
                - resources.log
        """
        self._log_directory = os.path.expanduser(log_directory)
        self.name = os.path.split(log_directory)[-1]
        self.phases: List[Tuple[str, str, datetime, float]] = []
        self._last_tell = 0
        self.evaluations: pd.DataFrame = pd.DataFrame()
        self.individuals: Dict[str, Individual] = dict()

        # Parse initialization/progress information from gama.log
        with open(os.path.join(log_directory, "gama.log")) as fh:
            gama_lines = fh.readlines()

        for line in gama_lines:
            if "INIT:" in line:
                self.hyperparameters = init_to_hps(line)
            if "STOP:" in line:
                time_and_place, activity = line.split(" STOP: ")
                timestamp, place = time_and_place[1:-1].split(" - ")
                end_time = datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S,%f")
                print(activity)
                phase, algorithm, _, duration = activity.split(" ")
                d = float(duration[:-3])  # e.g. "0.1300s.\n"
                start_time = end_time - timedelta(seconds=d)
                self.phases.append((phase, algorithm, start_time, d))

        self.metrics = self.hyperparameters["scoring"].split(",")
        if self.hyperparameters["regularize_length"]:
            self.metrics += ["length"]

        self.incomplete = len(self.phases) < 3
        self.update(force=True)  # updates self.evaluations and self.method_data
        # [ ] Dashboard -- how point to directory?? => #97

        self.search_method = self.hyperparameters["search"]

    def update(self, force: bool = False) -> bool:
        if not force and not self.incomplete:
            return False

        with open(os.path.join(self._log_directory, "evaluations.log"), "r") as fh:
            header = fh.readline()[:-1]
            self._last_tell = max(self._last_tell, fh.tell())
            fh.seek(self._last_tell)
            try:
                df = pd.read_csv(fh, sep=";", header=None, index_col=False)
            except pd.errors.EmptyDataError:
                return False
            self._last_tell = fh.tell()

            df.columns = header.split(";")
            df["n"] = df.index
            df = df.rename(columns=dict(t_start="start", t_wallclock="duration"))

            def tuple_to_metrics(tuple_str):
                return pd.Series([float(value) for value in tuple_str[1:-1].split(",")])

            df[self.metrics] = df.score.apply(tuple_to_metrics)
            df.start = pd.to_datetime(df.start)  # needed?
            df.duration = pd.to_timedelta(df.duration, unit="s")

            new_individuals = {
                id_: Individual.from_string(pipeline, pset)
                for id_, pipeline in zip(df.id, df.pipeline)
            }

            # Merge with previous records
            self.individuals.update(new_individuals)
            if self.evaluations.empty:
                self.evaluations = df
            else:
                df["n"] += self.evaluations.n.max() + 1
                self.evaluations = pd.concat([self.evaluations, df])
            df = self.evaluations

            search_start = df.start.min()
            for metric in self.metrics:
                df[f"{metric}_cummax"] = df[metric].cummax()
            if len(df.start) > 0:
                df["relative_end"] = (
                    (df.start + df.duration) - search_start
                ).dt.total_seconds()
            else:
                df["relative_end"] = pd.Series()
        return True

    @property
    def successful_evaluations(self):
        """ Return only evaluations that completed successfully """
        with pd.option_context("mode.use_inf_as_na", True):
            return self.evaluations[~self.evaluations[self.metrics].isna().any(axis=1)]


# new
def init_to_hps(init_line: str) -> Dict[str, str]:
    all_arguments = init_line.split("(", maxsplit=1)[-1].rsplit(")", maxsplit=1)[0]
    # A little hackery to get nested configurations (e.g. of search) to work
    # only supports one nested level - will do proper parsing later
    for token in ["()", "(", ")", ",,"]:
        all_arguments = all_arguments.replace(token, ",")
    return dict(hp.split("=") for hp in all_arguments.split(","))  # type: ignore



from typing import NamedTuple

DATA_TERMINAL = "data"


class Terminal(NamedTuple):
    """ Specifies a specific value for a specific type or input.

    E.g. a value for a hyperparameter for an algorithm.
    """

    value: object
    output: str
    identifier: str

    def __str__(self):
        """ str: e.g. "tol=0.5" """
        return f"{self.output}={format_hyperparameter_value(self.value)}"

    def __repr__(self):
        """ str: e.g. "FastICA.tol=0.5".

        If the hyperparameter is shared across primitives, there is no prefix.
        """
        return f"{self.identifier}={format_hyperparameter_value(self.value)}"


def format_hyperparameter_value(value: object) -> str:
    if isinstance(value, str):
        return f"'{value}'"  # Quoted
    elif callable(value) and hasattr(value, "__name__"):
        return f"{value.__name__}"  # type: ignore
    else:
        return str(value)



""" Implementation of NSGA-II and its subroutines as defined in Deb et al. (2002)

Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. A. M. T. (2002).
A fast and elitist multiobjective genetic algorithm: NSGA-II.
IEEE transactions on evolutionary computation, 6(2), 182-197.
"""
import itertools
import random
from functools import cmp_to_key
from typing import List, Any, Callable
import numpy as np


class NSGAMeta:
    def __init__(self, obj, metrics):
        self.obj = obj
        self.values = tuple((m(obj) for m in metrics))
        self.rank = None
        self.distance = 0
        self.dominating = []
        self.domination_counter = 0

    def dominates(self, other: "NSGAMeta"):
        for self_val, other_val in zip(self.values, other.values):
            if self_val <= other_val:  # or maybe <?
                return False
        return True

    def crowd_compare(self, other: "NSGAMeta"):
        """ Favor higher rank, if equal, favor less crowded. """
        self_better = self.rank < other.rank or (
            self.rank == other.rank and self.distance > other.distance
        )
        return -1 if self_better else 1


def nsga2_select(
    population: List[Any], n: int, metrics: List[Callable[[Any], float]]
) -> List[Any]:
    """ Select n pairs from the population.

     Selection is done through binary tournament selection based on crowding distance.
     Parent pairs may be repeated, but each parent pair consists of two unique parents.
     The population must be at least size 3 (otherwise it is trivial or impossible).
    """
    if len(population) < 3:
        raise ValueError("population must be at least size 3 for a pair to be selected")

    # Entire population is returned, but with rank and distance information.
    candidates = nsga2(population, n=len(population), metrics=metrics, return_meta=True)

    def select_one(exclude=None):
        selected = random.sample(candidates, k=3)
        ind1, ind2 = [s for s in selected if s != exclude][:2]
        return ind1 if ind1.crowd_compare(ind2) < 0 else ind2

    selected = []
    for _ in range(n):
        first = select_one()
        second = select_one(exclude=first)
        selected.append((first.obj, second.obj))
    return selected


def nsga2(
    population: List[Any],
    n: int,
    metrics: List[Callable[[Any], float]],
    return_meta: bool = False,
) -> List[Any]:
    """ Selects n individuals from the population for offspring according to NSGA-II.

    Parameters
    ----------
    population: List[T]
        A list of objects.
    n: int
        Number of objects to pick out of population.
        Must be greater than 0 and smaller than len(population).
    metrics: List[Callable[[T], float]]
        List of functions which obtain the values for each dimension
        on which to compare elements of population.
    return_meta: bool (default=False)
        If True, return the selected individuals wrapped in a NSGAMeta class,
        with information such as rank and distance.
        If False, return the selected individuals as they were passed to this function.

    Returns
    -------
    selection: List[T]
        A list of size n containing a subset of population.
    """
    if n == 0 or n > len(population):
        raise ValueError(f"n is {n} must be 0 < n < len(population)={len(population)}")
    population = [NSGAMeta(p, metrics) for p in population]
    selection: List[NSGAMeta] = []
    fronts = fast_non_dominated_sort(population)
    i = 0

    while len(selection) < n:
        crowding_distance_assignment(fronts[i])
        if len(selection) + len(fronts[i]) < n:
            # Enough space for the entire Pareto front, include all
            selection += fronts[i]
        else:
            # Only the least crowded remainder is selected
            s = sorted(fronts[i], key=cmp_to_key(lambda x, y: x.crowd_compare(y)))
            selection += s[: (n - len(selection))]  # Fill up to n
        i += 1

    return selection if return_meta else [s.obj for s in selection]


def fast_non_dominated_sort(P: List[NSGAMeta]) -> List[List[NSGAMeta]]:
    """ Sorts P into Pareto fronts. """
    fronts: List[List[NSGAMeta]] = [[]]
    for p, q in itertools.combinations(P, 2):
        if p.dominates(q):
            p.dominating.append(q)
            q.domination_counter += 1
        elif q.dominates(p):
            q.dominating.append(p)
            p.domination_counter += 1

    for p in P:
        if p.domination_counter == 0:
            p.rank = 1
            fronts[0].append(p)

    i = 0
    while fronts[i] != []:
        fronts.append([])
        for p in fronts[i]:
            for q in p.dominating:
                q.domination_counter -= 1
                if q.domination_counter == 0:
                    q.rank = i + 1
                    fronts[i + 1].append(q)
        i += 1
    return fronts


def crowding_distance_assignment(I: List[NSGAMeta]) -> None:
    for m in range(len(I[0].values)):
        I = sorted(I, key=lambda x: x.values[m])  # noqa: E741 'I' is name in paper
        I[0].distance = I[-1].distance = float("inf")
        if (
            I[-1].values[m] == I[0].values[m]
            or np.isinf(I[0].values[m])
            or np.isinf(I[-1].values[m])
        ):
            # Would raise divisionbyzero later, or give other numerical warnings.
            # This typically happens only for the worst pareto front(s),
            # so the inaccuracy in crowding distance for the remainder is not a concern.
            # Might consider immediately removing failing individuals.
            continue

        for i_prev, i, i_next in zip(I, I[1:], I[2:]):
            i.distance += (i_next.values[m] - i_prev.values[m]) / (
                I[-1].values[m] - I[0].values[m]
            )



import numpy as np

from sklearn.cluster import FeatureAgglomeration
from sklearn.preprocessing import (
    MaxAbsScaler,
    MinMaxScaler,
    Normalizer,
    PolynomialFeatures,
    RobustScaler,
    StandardScaler,
    Binarizer,
)
from sklearn.kernel_approximation import Nystroem, RBFSampler
from sklearn.decomposition import PCA, FastICA
from sklearn.feature_selection import (
    SelectFwe,
    SelectPercentile,
    VarianceThreshold,
    f_regression,
)


from sklearn.linear_model import ElasticNetCV, LassoLarsCV
from sklearn.ensemble import (
    ExtraTreesRegressor,
    GradientBoostingRegressor,
    AdaBoostRegressor,
    RandomForestRegressor,
)
from sklearn.tree import DecisionTreeRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import LinearSVR

# For comparison, this selection of operators and hyperparameters is
# currently most of what TPOT supports.

reg_config = {
    ElasticNetCV: {
        "l1_ratio": np.arange(0.0, 1.01, 0.05),
        "tol": [1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
    },
    ExtraTreesRegressor: {
        "n_estimators": [100],
        "max_features": np.arange(0.05, 1.01, 0.05),
        "min_samples_split": range(2, 21),
        "min_samples_leaf": range(1, 21),
        "bootstrap": [True, False],
    },
    GradientBoostingRegressor: {
        "n_estimators": [100],
        "loss": ["ls", "lad", "huber", "quantile"],
        "learning_rate": [1e-3, 1e-2, 1e-1, 0.5, 1.0],
        "max_depth": range(1, 11),
        "min_samples_split": range(2, 21),
        "min_samples_leaf": range(1, 21),
        "subsample": np.arange(0.05, 1.01, 0.05),
        "max_features": np.arange(0.05, 1.01, 0.05),
        "alpha": [0.75, 0.8, 0.85, 0.9, 0.95, 0.99],
    },
    AdaBoostRegressor: {
        "n_estimators": [100],
        "learning_rate": [1e-3, 1e-2, 1e-1, 0.5, 1.0],
        "loss": ["linear", "square", "exponential"],
        # 'max_depth': range(1, 11) not available in sklearn==0.19.1
    },
    DecisionTreeRegressor: {
        "max_depth": range(1, 11),
        "min_samples_split": range(2, 21),
        "min_samples_leaf": range(1, 21),
    },
    KNeighborsRegressor: {
        "n_neighbors": range(1, 101),
        "weights": ["uniform", "distance"],
        "p": [1, 2],
    },
    LassoLarsCV: {"normalize": [True, False]},
    LinearSVR: {
        "loss": ["epsilon_insensitive", "squared_epsilon_insensitive"],
        "dual": [True, False],
        "tol": [1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
        "C": [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],
        "epsilon": [1e-4, 1e-3, 1e-2, 1e-1, 1.0],
    },
    RandomForestRegressor: {
        "n_estimators": [100],
        "max_features": np.arange(0.05, 1.01, 0.05),
        "min_samples_split": range(2, 21),
        "min_samples_leaf": range(1, 21),
        "bootstrap": [True, False],
    },
    # Preprocesssors
    Binarizer: {"threshold": np.arange(0.0, 1.01, 0.05)},
    FastICA: {"tol": np.arange(0.0, 1.01, 0.05)},
    FeatureAgglomeration: {
        "linkage": ["ward", "complete", "average"],
        "affinity": ["euclidean", "l1", "l2", "manhattan", "cosine", "precomputed"],
        "param_check": [
            lambda params: (not params["linkage"] == "ward")
            or params["affinity"] == "euclidean"
        ],
    },
    MaxAbsScaler: {},
    MinMaxScaler: {},
    Normalizer: {"norm": ["l1", "l2", "max"]},
    Nystroem: {
        "kernel": [
            "rbf",
            "cosine",
            "chi2",
            "laplacian",
            "polynomial",
            "poly",
            "linear",
            "additive_chi2",
            "sigmoid",
        ],
        "gamma": np.arange(0.0, 1.01, 0.05),
        "n_components": range(1, 11),
    },
    PCA: {"svd_solver": ["randomized"], "iterated_power": range(1, 11)},
    PolynomialFeatures: {
        "degree": [2],
        "include_bias": [False],
        "interaction_only": [False],
    },
    RBFSampler: {"gamma": np.arange(0.0, 1.01, 0.05)},
    RobustScaler: {},
    StandardScaler: {},
    # Selectors
    SelectFwe: {"alpha": np.arange(0, 0.05, 0.001), "score_func": {f_regression: None}},
    SelectPercentile: {"percentile": range(1, 100), "score_func": {f_regression: None}},
    VarianceThreshold: {"threshold": np.arange(0.05, 1.01, 0.05)},
}



from typing import List, Optional

import pandas as pd
from plotly import graph_objects as go

from gama.logging.GamaReport import GamaReport


def plot_preset_graph(reports: List[GamaReport], preset: Optional[str]):
    if reports == []:
        return {}

    plots = []
    layout = {}
    first_metric = f"{reports[0].metrics[0]}"
    first_metric_max = f"{first_metric}_cummax"

    if preset == "best_over_n":
        # if aggregate == "separate-line":
        plots = [
            individual_plot(report, "n", first_metric_max, "lines")
            for report in reports
        ]
        # elif aggregate == "aggregate":
        #     plots = aggregate_plot(aggregate_df, "n", first_metric_max)
        layout = dict(
            title="Best score by iteration",
            xaxis=dict(title="n"),
            yaxis=dict(title=f"max {first_metric}"),
            hovermode="closest",
        )
    elif preset == "best_over_time":
        # if aggregate == "separate-line":
        plots = [
            individual_plot(report, "relative_end", first_metric_max, "lines")
            for report in reports
        ]
        # elif aggregate == "aggregate":
        #     plots = aggregate_best_over_time(aggregate_df, first_metric_max)
        layout = dict(
            title=f"Best score over time",
            xaxis=dict(title="time (s)"),
            yaxis=dict(title=f"max {first_metric}"),
            hovermode="closest",
        )
    elif preset == "size_vs_metric":
        # if aggregate == "separate-line":
        plots = [
            individual_plot(report, first_metric, "length", "markers")
            for report in reports
        ]
        # elif aggregate == "aggregate":
        #     plots = []
        #     for method in aggregate_df.search_method.unique():
        #         method_df = aggregate_df[aggregate_df.search_method == method]
        #         plots.append(
        #             go.Scatter(
        #                 x=method_df[first_metric],
        #                 y=method_df.length,
        #                 mode="markers",
        #                 name=method,
        #             )
        #         )
        layout = dict(
            title=f"Size vs {first_metric}",
            xaxis=dict(title=first_metric),
            yaxis=dict(title="pipeline length"),
            hovermode="closest",
        )
    elif preset == "number_pipeline_by_size":
        # if aggregate == "separate-line":
        for report in reports:
            size_counts = report.evaluations.length.value_counts()
            size_ratio = size_counts / len(report.individuals)
            plots.append(
                go.Bar(x=size_ratio.index.values, y=size_ratio.values, name=report.name)
            )
        # elif aggregate == "aggregate":
        #     for method in aggregate_df.search_method.unique():
        #        results_for_method = aggregate_df[aggregate_df.search_method == method]
        #        size_counts = results_for_method.length.value_counts()
        #        size_ratio = size_counts / len(results_for_method)
        #        plots.append(
        #            go.Bar(x=size_ratio.index.values, y=size_ratio.values, name=method)
        #        )
        layout = dict(
            title=f"Ratio of pipelines by size",
            xaxis=dict(title="pipeline length"),
            yaxis=dict(title="pipeline count"),
        )
    elif preset == "number_pipeline_by_learner":
        for report in reports:
            main_learners = [
                str(ind.main_node._primitive) for ind in report.individuals.values()
            ]
            learner_counts = pd.Series(main_learners).value_counts()
            learner_ratio = learner_counts / len(report.individuals)
            plots.append(
                go.Bar(
                    x=learner_ratio.index.values,
                    y=learner_ratio.values,
                    name=report.name,
                )
            )
        layout = dict(
            title=f"Ratio of pipelines by learner",
            xaxis=dict(title="pipeline length"),
            yaxis=dict(title="learner"),
        )
    elif preset == "evaluation_times_dist":
        # if aggregate == "separate-line":
        for report in reports:
            time_s = report.evaluations.duration.dt.total_seconds()
            plots.append(go.Histogram(x=time_s, name=report.name))
        # elif aggregate == "aggregate":
        #     for method in aggregate_df.search_method.unique():
        #         time_s = aggregate_df[
        #             aggregate_df.search_method == method
        #         ].duration.dt.total_seconds()
        #         plots.append(go.Histogram(x=time_s, name=method))
        layout = dict(
            title=f"Pipeline Evaluation Times",
            xaxis=dict(title="duration (s)"),
            yaxis=dict(title="count"),
        )
    elif preset == "n_by_rung":
        for report in reports:
            if report.search_method == "AsynchronousSuccessiveHalving":
                count_by_rung = (
                    report.evaluations.groupby(by="rung").n.count().reset_index()
                )
                plots.append(
                    go.Bar(x=count_by_rung.rung, y=count_by_rung.n, name=report.name)
                )
        layout = dict(
            title=f"#Evaluations by Rung",
            xaxis=dict(title="rung"),
            yaxis=dict(title="count"),
        )
    elif preset == "time_by_rung":
        for report in reports:
            if report.search_method == "AsynchronousSuccessiveHalving":
                duration_by_rung = (
                    report.evaluations.groupby(by="rung").duration.sum().reset_index()
                )
                duration_by_rung.duration = duration_by_rung.duration.dt.total_seconds()
                plots.append(
                    go.Bar(
                        x=duration_by_rung.rung,
                        y=duration_by_rung.duration,
                        name=report.name,
                    )
                )
        layout = dict(
            title=f"Time spent per Rung",
            xaxis=dict(title="rung"),
            yaxis=dict(title="time (s)"),
        )
    return {"data": plots, "layout": layout}


def individual_plot(report: GamaReport, x_axis: str, y_axis: str, mode: str):
    """

    :param report: report to pull data from
    :param x_axis: metric on the x-axis, column of report.evaluations
    :param y_axis: metric on the y-axis, column of report.evaluations
    :param mode: See `https://plot.ly/python/reference/#scatter-mode`
    :return:
        dash graph
    """
    print(len(report.evaluations[x_axis]))
    print(len(report.evaluations[y_axis]))
    return go.Scatter(
        name=f"{report.name}",
        x=report.evaluations[x_axis],
        y=report.evaluations[y_axis],
        text=[ind.short_name for ind in report.individuals.values()],
        mode=mode,
    )


def aggregate_best_over_time(aggregate: pd.DataFrame, y_axis: str):
    # By method, find the best score per trace at each point in time.
    # I'm sure there is a better way to do this...
    colors = {
        0: "rgba(255, 0, 0, {a})",
        1: "rgba(0, 255, 0, {a})",
        2: "rgba(0, 0, 255, {a})",
    }
    soft_colors = {i: color.format(a=0.2) for i, color in colors.items()}
    hard_colors = {i: color.format(a=1.0) for i, color in colors.items()}

    aggregate_data: List[go.Scatter] = []
    aggregate = aggregate[aggregate[y_axis] != -float("inf")]
    for color_no, method in enumerate(aggregate.search_method.unique()):
        method_agg = aggregate[aggregate.search_method == method]
        all_end_times = method_agg.relative_end.unique()
        end_time_series = pd.Series(sorted(all_end_times), name="relative_end")

        scores_over_time = pd.DataFrame(end_time_series)
        for log_no in method_agg.log_no.unique():
            log_results = method_agg[method_agg.log_no == log_no]
            best_for_log = pd.merge_asof(
                end_time_series,
                log_results[["relative_end", y_axis]],
                on="relative_end",
            )
            scores_over_time[f"log{log_no}"] = best_for_log[y_axis]
        scores_over_time = scores_over_time.set_index("relative_end")
        mean_scores = scores_over_time.mean(axis=1)
        std_scores = scores_over_time.std(axis=1)
        scores_over_time["mean"] = mean_scores
        scores_over_time["std"] = std_scores

        upper_bound = go.Scatter(
            x=scores_over_time.index,
            y=scores_over_time["mean"] + scores_over_time["std"],
            mode="lines",
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            fillcolor=soft_colors[color_no],
            fill="tonexty",
            showlegend=False,
        )

        mean_performance = go.Scatter(
            name=method,
            x=scores_over_time.index,
            y=scores_over_time["mean"],
            mode="lines",
            line=dict(color=hard_colors[color_no]),
            fillcolor=soft_colors[color_no],
            fill="tonexty",
        )

        lower_bound = go.Scatter(
            x=scores_over_time.index,
            y=scores_over_time["mean"] - scores_over_time["std"],
            mode="lines",
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            showlegend=False,
        )
        aggregate_data += [lower_bound, mean_performance, upper_bound]
    return aggregate_data


def aggregate_plot(aggregate: pd.DataFrame, x_axis: str, y_axis: str):
    """ Create an aggregate plot over multiple reports.

     Aggregates the mean and std of `y_axis` by `x_axis`.

    :param aggregate: dataframe with all evaluations
    :param x_axis: column which is grouped by before aggregating `y_axis`
    :param y_axis: column over which to calculate the mean/std.
    :return:
        Three dash Scatter objects which respectively draw  the lower bound, mean and
        upper bound.
    """
    colors = {
        0: "rgba(255, 0, 0, {a})",
        1: "rgba(0, 255, 0, {a})",
        2: "rgba(0, 0, 255, {a})",
    }
    soft_colors = {i: color.format(a=0.2) for i, color in colors.items()}
    hard_colors = {i: color.format(a=1.0) for i, color in colors.items()}

    # concat_df = pd.concat([report.evaluations for report in reports_to_combine])
    # concat_df = concat_df[concat_df[y_axis] != -float('inf')]
    # agg_df = concat_df.groupby(by=x_axis).agg({y_axis: ['mean', 'std']}).reset_index()
    # agg_df.columns = [x_axis, y_axis, 'std']
    aggregate_data: List[go.Scatter] = []
    aggregate = aggregate[aggregate[y_axis] != -float("inf")]
    for color_no, method in enumerate(aggregate.search_method.unique()):
        agg_for_method = aggregate[aggregate.search_method == method]
        agg_df = (
            agg_for_method.groupby(by=x_axis)
            .agg({y_axis: ["mean", "std"]})
            .reset_index()
        )
        agg_df.columns = [x_axis, y_axis, "std"]

        upper_bound = go.Scatter(
            x=agg_df[x_axis],
            y=agg_df[y_axis] + agg_df["std"],
            mode="lines",
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            fillcolor=soft_colors[color_no],
            fill="tonexty",
            showlegend=False,
        )

        mean_performance = go.Scatter(
            name=method,
            x=agg_df[x_axis],
            y=agg_df[y_axis],
            mode="lines",
            line=dict(color=hard_colors[color_no]),
            fillcolor=soft_colors[color_no],
            fill="tonexty",
        )

        lower_bound = go.Scatter(
            x=agg_df[x_axis],
            y=agg_df[y_axis] - agg_df["std"],
            mode="lines",
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            showlegend=False,
        )
        aggregate_data += [lower_bound, mean_performance, upper_bound]
    return aggregate_data



""" Selection operators. """
import random
from typing import List

from gama.genetic_programming.operator_set import OperatorSet
from gama.genetic_programming.components import Individual
from gama.genetic_programming.nsga2 import nsga2_select
from gama.utilities.generic.paretofront import ParetoFront
from gama.genetic_programming.crossover import _valid_crossover_functions


def create_from_population(
    operator_shell: OperatorSet,
    pop: List[Individual],
    n: int,
    cxpb: float,
    mutpb: float,
) -> List[Individual]:
    """ Creates n new individuals based on the population. """
    offspring = []
    metrics = [lambda ind: ind.fitness.values[0], lambda ind: ind.fitness.values[1]]
    parent_pairs = nsga2_select(pop, n, metrics)
    for (ind1, ind2) in parent_pairs:
        if random.random() < cxpb and len(_valid_crossover_functions(ind1, ind2)) > 0:
            ind1 = operator_shell.mate(ind1, ind2)
        else:
            ind1 = operator_shell.mutate(ind1)
        offspring.append(ind1)
    return offspring


def eliminate_from_pareto(pop: List[Individual], n: int) -> List[Individual]:
    # For now we only eliminate one at a time so this will do.
    if n != 1:
        raise NotImplementedError("Currently only n=1 is supported.")

    def inverse_fitness(ind):
        return [-value for value in ind.fitness.values]

    pareto_worst = ParetoFront(pop, inverse_fitness)
    return [random.choice(pareto_worst)]



import copy
from typing import Tuple, List, Set

from sklearn.base import TransformerMixin
from gama.genetic_programming.components import Individual


def transformers_to_str(transformers: List[TransformerMixin]) -> List[str]:
    """ Format a transformer for code export, removes any mapping. """
    copies = list(map(copy.copy, transformers))
    for transformer in copies:
        if hasattr(transformer, "mapping"):
            transformer.mapping = None  # type: ignore  # ignore no attr 'mapping'
    return list(map(str, copies))


def format_import(o: object) -> str:
    """ Creates the import statement for `o`'s class. """
    if o.__module__.split(".")[-1].startswith("_"):
        module = ".".join(o.__module__.split(".")[:-1])
    else:
        module = o.__module__
    return f"from {module} import {o.__class__.__name__}"


def format_pipeline(steps: List[Tuple[str, str]], name: str = "pipeline"):
    steps_str = ",\n".join([f"('{name}', {step})" for name, step in steps])
    return f"{name} = Pipeline([{steps_str}])\n"


def imports_and_steps_for_individual(
    individual: Individual,
) -> Tuple[Set[str], List[Tuple[str, str]]]:
    """ Determine required imports and steps for the individual's pipeline.

    Returns two lists:
     - one with import statements
     - one with tuples with pipeline step names and step values

    E.g. (["from sklearn.naive_bayes import GaussianNB"], [('0', 'GaussianNB()')])
    """
    imports = ["from numpy import nan", "from sklearn.pipeline import Pipeline"]
    imports += [format_import(step) for name, step in individual.pipeline.steps]

    steps = []
    for i, primitive_node in reversed(list(enumerate(individual.primitives))):
        steps.append((str(i), primitive_node.str_nonrecursive))
        for terminal in primitive_node._terminals:
            if callable(terminal.value) and hasattr(terminal.value, "__name__"):
                imports.append(
                    f"from {terminal.value.__module__} import {terminal.value.__name__}"  # type: ignore # noqa: E501
                )

    return set(imports), steps


def individual_to_python(
    individual: Individual, prepend_steps: List[Tuple[str, TransformerMixin]] = None
) -> str:
    """ Generate code for the machine learning pipeline represented by `individual`. """
    imports, steps = imports_and_steps_for_individual(individual)
    if prepend_steps is not None:
        steps = prepend_steps + steps
        imports = imports.union({format_import(step) for _, step in prepend_steps})
    steps_str = ",\n".join([f"('{name}', {step})" for name, step in steps])
    pipeline = f"Pipeline([{steps_str}])"
    script = "\n".join(sorted(imports)) + "\n\n" + "pipeline = " + pipeline + "\n"

    return script



""" Provides updates on GAMA's search.
 Next step is to call GAMA directly, but the `fit` call has to be made async. """

import shlex
import subprocess
from collections import defaultdict


class Controller:
    def __init__(self):
        self._subscribers = defaultdict(list)

    def start_gama(
        self,
        metric,
        regularize,
        n_jobs,
        max_total_time_h,
        max_total_time_m,
        max_eval_time_h,
        max_eval_time_m,
        input_file,
        log_dir,
        target,
    ):
        # For some reason, 0 input registers as None.
        max_total_time_h = 0 if max_total_time_h is None else max_total_time_h
        max_total_time_m = 0 if max_total_time_m is None else max_total_time_m
        max_eval_time_h = 0 if max_eval_time_h is None else max_eval_time_h
        max_eval_time_m = 0 if max_eval_time_m is None else max_eval_time_m
        max_total_time = max_total_time_h * 60 + max_total_time_m
        max_eval_time = max_eval_time_h * 60 + max_eval_time_m
        command = (
            f'gama "{input_file}" -v -n {n_jobs} -t {max_total_time} '
            f'--time_pipeline {max_eval_time} -outdir {log_dir} --target "{target}"'
        )
        if regularize != "on":
            command += " --long"
        if metric != "default":
            command += f" -m {metric}"

        command = shlex.split(command)
        # fake_command = ['python', '-h']
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self._on_gama_started(process, log_dir)

    def _on_gama_started(self, process, log_file):
        for subscriber in self._subscribers["gama_started"]:
            subscriber(process, log_file)

    def gama_started(self, callback_function):
        self._subscribers["gama_started"].append(callback_function)

    def gama_ended(self, callback_function):
        self._subscribers["gama_ended"].append(callback_function)



import logging

from .components import Individual

log = logging.getLogger(__name__)


class OperatorSet:
    """ Provides a thin layer for ea operators for logging, callbacks and safety. """

    def __init__(
        self,
        mutate,
        mate,
        create_from_population,
        create_new,
        compile_,
        eliminate,
        evaluate_callback,
        max_retry=50,
        completed_evaluations=None,
    ):
        """

        :param mutate:
        :param mate:
        :param create:
        :param create_new:
        """

        self._mutate = mutate
        self._mate = mate
        self._create_from_population = create_from_population
        self._create_new = create_new
        self._compile = compile_
        self._safe_compile = None
        self._eliminate = eliminate
        self._max_retry = max_retry
        self._evaluate = None
        self._evaluate_callback = evaluate_callback
        self.evaluate = None

        self._completed_evaluations = completed_evaluations

    def wait_next(self, async_evaluator):
        future = async_evaluator.wait_next()
        if future.result is not None:
            evaluation = future.result
            if self._evaluate_callback is not None:
                self._evaluate_callback(evaluation)

        elif future.exception is not None:
            log.warning(f"Error raised during evaluation: {str(future.exception)}.")
        return future

    def try_until_new(self, operator, *args, **kwargs):
        for _ in range(self._max_retry):
            individual = operator(*args, **kwargs)
            if str(individual.main_node) not in self._completed_evaluations:
                return individual
        else:
            log.debug(f"50 iterations of {operator.__name__} did not yield new ind.")
            # For progress on solving this, see #11
            return individual

    def mate(self, ind1: Individual, ind2: Individual, *args, **kwargs):
        def mate_with_log():
            new_individual1, new_individual2 = ind1.copy_as_new(), ind2.copy_as_new()
            self._mate(new_individual1, new_individual2, *args, **kwargs)
            new_individual1.meta = dict(parents=[ind1._id, ind2._id], origin="cx")
            return new_individual1

        individual = self.try_until_new(mate_with_log)
        return individual

    def mutate(self, ind: Individual, *args, **kwargs):
        def mutate_with_log():
            new_individual = ind.copy_as_new()
            mutator = self._mutate(new_individual, *args, **kwargs)
            new_individual.meta = dict(parents=[ind._id], origin=mutator.__name__)
            return new_individual

        ind = self.try_until_new(mutate_with_log)
        return ind

    def individual(self, *args, **kwargs):
        expression = self._create_new(*args, **kwargs)
        if self._safe_compile is not None:
            compile_ = self._safe_compile
        else:
            compile_ = self._compile
        ind = Individual(expression, to_pipeline=compile_)
        ind.meta["origin"] = "new"
        return ind

    def create(self, *args, **kwargs):
        return self._create_from_population(self, *args, **kwargs)

    def eliminate(self, *args, **kwargs):
        return self._eliminate(*args, **kwargs)



from typing import List, Optional, Sequence, Tuple

import pandas as pd
from sklearn.base import TransformerMixin

from gama.genetic_programming.components import Individual
from gama.postprocessing.base_post_processing import BasePostProcessing
from gama.utilities.export import (
    imports_and_steps_for_individual,
    transformers_to_str,
    format_import,
    format_pipeline,
)


class BestFitPostProcessing(BasePostProcessing):
    """ Post processing technique which trains the best found single pipeline. """

    def __init__(self, time_fraction: float = 0.1):
        super().__init__(time_fraction)
        self._selected_individual: Optional[Individual] = None

    def post_process(
        self, x: pd.DataFrame, y: pd.Series, timeout: float, selection: List[Individual]
    ) -> object:
        self._selected_individual = selection[0]
        return self._selected_individual.pipeline.fit(x, y)

    def to_code(
        self, preprocessing: Sequence[Tuple[str, TransformerMixin]] = None
    ) -> str:
        if self._selected_individual is None:
            raise RuntimeError("`to_code` can only be called after `post_process`.")

        imports, steps = imports_and_steps_for_individual(self._selected_individual)
        if preprocessing is not None:
            trans_strs = transformers_to_str([t for _, t in preprocessing])
            names = [name for name, _ in preprocessing]
            steps = list(zip(names, trans_strs)) + steps
            imports = imports.union({format_import(t) for _, t in preprocessing})

        pipeline_statement = format_pipeline(steps)
        script = "\n".join(imports) + "\n\n" + pipeline_statement
        return script



from datetime import datetime
import logging
import os
import time
from typing import Callable, Tuple, Optional, Sequence

import stopit
from sklearn.base import TransformerMixin, is_classifier
from sklearn.model_selection import ShuffleSplit, cross_validate, check_cv
from sklearn.pipeline import Pipeline

from gama.utilities.evaluation_library import Evaluation
from gama.utilities.generic.stopwatch import Stopwatch
import numpy as np
from gama.utilities.metrics import Metric
from gama.genetic_programming.components import Individual, PrimitiveNode, Fitness

log = logging.getLogger(__name__)


def primitive_node_to_sklearn(primitive_node: PrimitiveNode) -> object:
    hyperparameters = {
        terminal.output: terminal.value for terminal in primitive_node._terminals
    }
    return primitive_node._primitive.identifier(**hyperparameters)


def compile_individual(
    individual: Individual,
    parameter_checks=None,
    preprocessing_steps: Sequence[Tuple[str, TransformerMixin]] = None,
) -> Pipeline:
    steps = [
        (str(i), primitive_node_to_sklearn(primitive))
        for i, primitive in enumerate(individual.primitives)
    ]
    if preprocessing_steps:
        steps = steps + list(reversed(preprocessing_steps))
    return Pipeline(list(reversed(steps)))


def object_is_valid_pipeline(o):
    """ Determines if object behaves like a scikit-learn pipeline. """
    return (
        o is not None
        and hasattr(o, "fit")
        and hasattr(o, "predict")
        and hasattr(o, "steps")
    )


def evaluate_pipeline(
    pipeline, x, y_train, timeout: float, metrics: Tuple[Metric], cv=5, subsample=None,
) -> Tuple:
    """ Score `pipeline` with k-fold CV according to `metrics` on (a subsample of) X, y

    Returns
    -------
    Tuple:
        prediction: np.ndarray if successful, None if not
        scores: tuple with one float per metric, each value is -inf on fail.
        estimators: list of fitted pipelines if successful, None if not
        error: None if successful, otherwise an Exception
    """
    if not object_is_valid_pipeline(pipeline):
        raise TypeError(f"Pipeline must not be None and requires fit, predict, steps.")

    prediction, estimators = None, None
    # default score for e.g. timeout or failure
    scores = tuple([float("-inf")] * len(metrics))

    with stopit.ThreadingTimeout(timeout) as c_mgr:
        try:
            if isinstance(subsample, int) and subsample < len(y_train):
                sampler = ShuffleSplit(n_splits=1, train_size=subsample, random_state=0)
                idx, _ = next(sampler.split(x))
                x, y_train = x.iloc[idx, :], y_train[idx]

            splitter = check_cv(cv, y_train, is_classifier(pipeline))
            result = cross_validate(
                pipeline,
                x,
                y_train,
                cv=splitter,
                return_estimator=True,
                scoring=[m.name for m in metrics],
                error_score="raise",
            )
            scores = tuple([np.mean(result[f"test_{m.name}"]) for m in metrics])
            estimators = result["estimator"]

            for (estimator, (_, test)) in zip(estimators, splitter.split(x, y_train)):
                if any([m.requires_probabilities for m in metrics]):
                    fold_pred = estimator.predict_proba(x.iloc[test, :])
                else:
                    fold_pred = estimator.predict(x.iloc[test, :])

                if prediction is None:
                    if fold_pred.ndim == 2:
                        prediction = np.empty(shape=(len(y_train), fold_pred.shape[1]))
                    else:
                        prediction = np.empty(shape=(len(y_train),))
                prediction[test] = fold_pred

            # prediction, scores, estimators = cross_val_predict_score(
            #     pipeline, x, y_train, cv=cv, metrics=metrics
            # )
        except stopit.TimeoutException:
            # This exception is handled by the ThreadingTimeout context manager.
            raise
        except KeyboardInterrupt:
            raise
        except Exception as e:
            return prediction, scores, estimators, e

    if c_mgr.state == c_mgr.INTERRUPTED:
        # A TimeoutException was raised, but not by the context manager.
        # This indicates that the outer context manager (the ea) timed out.
        raise stopit.utils.TimeoutException()

    if not c_mgr:
        # For now we treat an eval timeout the same way as
        # e.g. NaN exceptions and use the default score.
        return prediction, scores, estimators, stopit.TimeoutException()

    return prediction, tuple(scores), estimators, None


def evaluate_individual(
    individual: Individual,
    evaluate_pipeline: Callable,
    timeout: float = 1e6,
    deadline: Optional[float] = None,
    add_length_to_score: bool = True,
    **kwargs,
) -> Evaluation:
    """ Evaluate the pipeline specified by individual, and record

    Parameters
    ----------
    individual: Individual
        Blueprint for the pipeline to evaluate.
    evaluate_pipeline: Callable
        Function which takes the pipeline and produces validation predictions,
        scores, estimators and errors.
    timeout: float (default=1e6)
        Maximum time in seconds that the evaluation is allowed to take.
        Don't depend on high accuracy.
        A shorter timeout is imposed if `deadline` is in less than `timeout` seconds.
    deadline: float, optional
        A time in seconds since epoch.
        Cut off evaluation at `deadline` even if `timeout` seconds have not yet elapsed.
    add_length_to_score: bool (default=True)
        Add the length of the individual to the score result of the evaluation.
    **kwargs: Dict, optional (default=None)
        Passed to `evaluate_pipeline` function.

    Returns
    -------
    Evaluation

    """
    result = Evaluation(individual, pid=os.getpid())
    result.start_time = datetime.now()

    if deadline is not None:
        time_to_deadline = deadline - time.time()
        timeout = min(timeout, time_to_deadline)

    with Stopwatch() as wall_time, Stopwatch(time.process_time) as process_time:
        evaluation = evaluate_pipeline(individual.pipeline, timeout=timeout, **kwargs)
        result._predictions, result.score, result._estimators, error = evaluation
        if error is not None:
            result.error = f"{type(error)} {str(error)}"
    result.duration = wall_time.elapsed_time

    if add_length_to_score:
        result.score = result.score + (-len(individual.primitives),)
    individual.fitness = Fitness(
        result.score,
        result.start_time,
        wall_time.elapsed_time,
        process_time.elapsed_time,
    )

    return result



from typing import Union, Type, Tuple

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype

from gama.utilities.preprocessing import log


def series_looks_categorical(series) -> bool:
    if series.dtype == "object":
        return True
    if is_numeric_dtype(series):
        value_counts = series.value_counts()
        integer_like = series.dtype.kind == "i" or all(
            x.is_integer() for x in series.dropna()
        )
        return len(value_counts) <= 10 and integer_like
    return False


def infer_categoricals_inplace(df):
    """ Use simple heuristics to guess which columns should be categorical. """
    for column in df:
        if series_looks_categorical(df[column]):
            df[column] = df[column].astype("category")


def numpy_to_dataframe(x: np.ndarray) -> pd.DataFrame:
    x = pd.DataFrame(x).infer_objects()
    x = x.infer_objects()
    infer_categoricals_inplace(x)
    return x


def format_y(
    y: Union[pd.DataFrame, pd.Series, np.ndarray], y_type: Type = pd.Series
) -> Union[pd.DataFrame, pd.Series]:
    """ Transforms a target vector or indicator matrix to a single series (or 1d df) """
    if not isinstance(y, (np.ndarray, pd.Series, pd.DataFrame)):
        raise TypeError("y must be np.ndarray, pd.Series or pd.DataFrame.")
    if y_type not in [pd.Series, pd.DataFrame]:
        raise ValueError(f"`y_type` must be pd.Series or pd.DataFrame but is {y_type}.")

    if isinstance(y, np.ndarray) and y.ndim == 2:
        # Either indicator matrix or should be a vector.
        if y.shape[1] > 1:
            y = np.argmax(y, axis=1)
        else:
            y = y.squeeze()

    if y_type == pd.Series:
        if isinstance(y, pd.DataFrame):
            y = y.squeeze()
        elif isinstance(y, np.ndarray):
            y = pd.Series(y)
        if series_looks_categorical(y):
            y = y.astype("category")
    elif y_type == pd.DataFrame:
        if not isinstance(y, pd.DataFrame):
            y = pd.DataFrame(y)
    return y


def remove_unlabeled_rows(
    x: pd.DataFrame, y: Union[pd.Series, pd.DataFrame]
) -> Tuple[pd.DataFrame, Union[pd.Series, pd.DataFrame]]:
    """ Removes all rows from x and y where y is nan. """
    if isinstance(y, pd.DataFrame):
        unlabeled = y.iloc[:, 0].isnull()
    else:
        unlabeled = y.isnull()

    if unlabeled.any():
        log.info(
            f"Target vector has been found to contain {sum(unlabeled)} NaN-labels, "
            f"these rows will be ignored."
        )
        x, y = x.loc[~unlabeled], y.loc[~unlabeled]
    return x, y


def format_x_y(
    x: Union[pd.DataFrame, np.ndarray],
    y: Union[pd.DataFrame, pd.Series, np.ndarray],
    y_type: Type = pd.Series,
    remove_unlabeled: bool = True,
) -> Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]:
    """ Take (X,y) data and convert it to a (pd.DataFrame, pd.Series) tuple.

    Parameters
    ----------
    x: pandas.DataFrame or numpy.ndarray
    y: pandas.DataFrame, pandas.Series or numpy.ndarray
    y_type: Type (default=pandas.Series)
    remove_unlabeled: bool (default=True)
        If true, remove all rows associated with unlabeled data (NaN in y).

    Returns
    -------
    Tuple[pandas.DataFrame, pandas.DataFrame or pandas.Series]
        X and y, where X is formatted as pd.DataFrame and y is formatted as `y_type`.
    """
    if not isinstance(x, (np.ndarray, pd.DataFrame)):
        raise TypeError("X must be either np.ndarray or pd.DataFrame.")

    if isinstance(x, np.ndarray):
        x = numpy_to_dataframe(x)
    if not isinstance(y, y_type):
        y = format_y(y, y_type)

    if remove_unlabeled:
        x, y = remove_unlabeled_rows(x, y)

    return x, y



""" This module contains functions for loading data. """
from collections import OrderedDict
import csv
from typing import Tuple, Optional, Dict, List

import arff
import pandas as pd

from gama.data_formatting import infer_categoricals_inplace

CSV_SNIFF_SIZE = 2 ** 12


def sniff_csv_meta(file_path: str) -> Tuple[str, bool]:
    """ Determine the csv delimiter and whether it has a header """
    with open(file_path, "r") as csv_file:
        first_bytes = csv_file.read(2 ** 12)
    sep = csv.Sniffer().sniff(first_bytes).delimiter
    has_header = csv.Sniffer().has_header(first_bytes)
    return sep, has_header


def load_csv_header(file_path: str, **kwargs) -> List[str]:
    """ Return column names in the header, or 0...N if no header is present. """
    if not file_path.endswith(".csv"):
        raise ValueError(f"{file_path} is not a file with .csv extension.")
    sep, has_header = sniff_csv_meta(file_path)
    sep = kwargs.get("sep", sep)

    with open(file_path, "r") as csv_file:
        first_line = csv_file.readline()[:-1]

    if has_header:
        return first_line.split(sep)
    else:
        return [str(i) for i, _ in enumerate(first_line.split(sep))]


def csv_to_pandas(file_path: str, **kwargs) -> pd.DataFrame:
    """ Load data from the csv file into a pd.DataFrame.

    Parameters
    ----------
    file_path: str
        Path of the csv file
    kwargs:
        Additional arguments for pandas.read_csv.
        If not specified, the presence of the header and the delimiter token are
        both detected automatically.

    Returns
    -------
    pandas.DataFrame
        A dataframe of the data in the ARFF file,
        with categorical columns having category dtype.
    """
    if "header" not in kwargs or "sep" not in kwargs:
        sep, has_header = sniff_csv_meta(file_path)
        kwargs["sep"] = kwargs.get("sep", sep)
        kwargs["header"] = kwargs.get("header", 0 if has_header else None)

    df = pd.read_csv(file_path, **kwargs).infer_objects()
    # Since CSV files do not have type annotation, we must infer their type to
    # know which preprocessing steps to apply.
    infer_categoricals_inplace(df)
    return df


def arff_to_pandas(
    file_path: str, encoding: Optional[str] = None, **kwargs
) -> pd.DataFrame:
    """ Load data from the ARFF file into a pd.DataFrame.

    Parameters
    ----------
    file_path: str
        Path of the ARFF file
    encoding: str, optional
        Encoding of the ARFF file.
    **kwargs:
        Any arugments for arff.load.

    Returns
    -------
    pandas.DataFrame
        A dataframe of the data in the ARFF file,
        with categorical columns having category dtype.
    """
    with open(file_path, "r", encoding=encoding) as arff_file:
        arff_dict = arff.load(arff_file, **kwargs)

    attribute_names, data_types = zip(*arff_dict["attributes"])
    data = pd.DataFrame(arff_dict["data"], columns=attribute_names)
    for attribute_name, dtype in arff_dict["attributes"]:
        # 'real' and 'numeric' are probably interpreted correctly.
        # Date support needs to be added.
        if isinstance(dtype, list):
            data[attribute_name] = data[attribute_name].astype("category")
    return data


def file_to_pandas(
    file_path: str, encoding: Optional[str] = None, **kwargs
) -> pd.DataFrame:
    """ Load ARFF/csv file into pd.DataFrame.

    Parameters
    ----------
    file_path: str
        path to the csv or ARFF file.
    encoding: str, optional
        Encoding, only used for ARFF files.
    kwargs:
        Any arguments for arff.load or pandas.read_csv

    Returns
    -------
    pd.DataFrame
    """
    if file_path.endswith(".arff"):
        data = arff_to_pandas(file_path, encoding, **kwargs)
    elif file_path.endswith(".csv"):
        data = csv_to_pandas(file_path, **kwargs)
    else:
        raise ValueError("Only csv and arff files supported.")
    return data


def X_y_from_file(
    file_path: str,
    split_column: Optional[str] = None,
    encoding: Optional[str] = None,
    **kwargs,
) -> Tuple[pd.DataFrame, pd.Series]:
    """ Load ARFF/csv file into pd.DataFrame and specified column to pd.Series.

    Parameters
    ----------
    file_path: str
        path to the ARFF file.
    split_column: str, optional (default=None)
        Column to split and return separately (e.g. target column).
        Value should either match a column name or None.
        If None is specified, the last column is returned separately.
    encoding: str, optional
        Encoding, only used for ARFF files.
    kwargs:
        Any arguments for arff.load or pandas.read_csv

    Returns
    -------
    Tuple[pd.DataFrame, pd.Series]
        Features (everything except split_column) and targets (split_column).
    """
    data = file_to_pandas(file_path, encoding, **kwargs)
    if split_column is None:
        return data.iloc[:, :-1], data.iloc[:, -1]
    elif split_column in data.columns:
        return data.loc[:, data.columns != split_column], data.loc[:, split_column]
    else:
        raise ValueError(f"No column named {split_column} found in {file_path}")


def load_feature_metadata_from_file(file_path: str) -> Dict[str, str]:
    """ Load the header of the csv or ARFF file, return the type of each attribute.

    For csv files, presence of a header is detected with the Python csv parser.
    If no header is present in the csv file, the columns will be labeled with a number.
    Additionally, the column types is not inferred for csv files.
    """
    if file_path.lower().endswith(".arff"):
        return load_feature_metadata_from_arff(file_path)
    elif file_path.lower().endswith(".csv"):
        return {c: "" for c in load_csv_header(file_path)}
    else:
        raise ValueError("Only csv and arff files supported.")


def load_feature_metadata_from_arff(file_path: str) -> Dict[str, str]:
    """ Load the header of the ARFF file and return the type of each attribute. """
    data_header = "@data"
    attribute_indicator = "@attribute"
    attributes = OrderedDict()
    with open(file_path, "r") as fh:
        line = fh.readline()
        while not line.lower().startswith(data_header):
            if line.lower().startswith(attribute_indicator):
                # arff uses a space separator, but allows spaces in
                # feature name (name must be quoted) and feature type (if nominal).
                indicator, name_and_type = line.split(" ", 1)
                if name_and_type.startswith('"'):
                    name, data_type = name_and_type[1:].split('" ', 1)
                    name = name
                else:
                    name, data_type = name_and_type.split(" ", 1)
                attributes[name] = data_type
            line = fh.readline()[:-1]  # remove newline character
    return attributes



from typing import List

try:
    import dash
    import dash_core_components as dcc
    import dash_bootstrap_components as dbc
    import dash_html_components as html
    from dash.dependencies import Input, Output
except ImportError:
    print(
        "Unable to import Dash modules. "
        "Make sure GAMA is installed with the optional visualization dependencies."
        "E.g.: pip install gama[vis]"
    )
    raise

from gama.dashboard.controller import Controller
from gama.dashboard.pages.base_page import BasePage


external_scripts = [
    "https://code.jquery.com/jquery-3.3.1.min.js",
    "https://cdn.datatables.net/v/dt/dt-1.10.18/datatables.min.js",
]
dashboard = dash.Dash(
    "GamaDashboard",
    external_stylesheets=[dbc.themes.BOOTSTRAP],
    external_scripts=external_scripts,
)
dashboard.config.suppress_callback_exceptions = True

controller = Controller()


# === Construct UI elements ===


def build_app():
    from gama.dashboard.pages import pages

    base = create_generic_layout()
    base["tabs"].children = create_tabs(pages)
    page_stores = []
    for page in pages:
        page.build_page(dashboard, controller)
        if hasattr(page, "gama_started"):
            controller.gama_started(page.gama_started)
        page_stores.append(
            dcc.Store(id=f"{page.id}-store", storage_type="session", data={})
        )
    base["page-stores"].children = page_stores
    return base


def create_generic_layout():
    """ Creates the generic layout of tabs and their content pages. """
    tab_banner_style = {
        "border-top-left-radius": "3px",
        "background-color": "#f9f9f9",
        "padding": "0px 24px",
        "border-bottom": "1px solid #d6d6d6",
    }

    return html.Div(
        id="page",
        children=[
            html.Div(id="tabs", style=tab_banner_style),
            html.Div(id="content"),
            html.Div(id="page-stores"),
        ],
        style={"font-family": "'Open Sans Semi Bold', sans-serif"},
    )


def create_tabs(pages: List[BasePage]):
    if pages == []:
        raise ValueError("Must have at least one tab.")

    # Sort pages by alignment
    sorted_pages = sorted(pages, key=lambda p: p.alignment)
    left_aligned = [page for page in sorted_pages if page.alignment >= 0]
    right_aligned = [page for page in sorted_pages if page.alignment < 0]
    grouped_pages = left_aligned + right_aligned
    tabs = [create_tab(page.name, page.starts_hidden) for page in grouped_pages]
    return [dcc.Tabs(id="page-tabs", value=tabs[0].value, children=tabs)]


def create_tab(name: str, invisible: bool):
    tab_style = {
        "color": "black",
        "width": "10%",
        "border-top-left-radius": "3px",  # round tab corners
        "border-top-right-radius": "3px",
        "border-bottom": "0px",  # bottom box-shadow still present
        "padding": "6px",
        "display": "none" if invisible else "inline",
    }
    selected_tab_style = {
        **tab_style,
        "border-top": "3px solid #c81818",  # Highlight color (TU/e colored)
        "box-shadow": "1px 1px 0px white",  # removes bottom edge
        "display": "inline",
    }
    return dcc.Tab(
        id=f"{name}-tab",
        label=name,
        value=name,
        style=tab_style,
        selected_style=selected_tab_style,
    )


# === Callbacks ===


@dashboard.callback([Output("content", "children")], [Input("page-tabs", "value")])
def display_page_content(page_name):
    from gama.dashboard.pages import pages

    page = [page for page in pages if page.name == page_name][0]
    page.need_update = True
    return [page.content]


def main(open_window: bool = True, **server_kwargs):
    dashboard.layout = build_app()
    # if open_window:
    #     import webbrowser
    #     from threading import Timer
    #
    #     port = server_kwargs.get("port", 8050)
    #     localhost = f"http://127.0.0.1:{port}"
    #
    #     Timer(1, webbrowser.open_new, kwargs=dict(url=localhost)).start()
    dashboard.run_server(**server_kwargs)


if __name__ == "__main__":
    main(debug=True)



from enum import Enum
from typing import Iterable, Tuple, Union

from sklearn.metrics import get_scorer
from sklearn.metrics._scorer import _ProbaScorer, _BaseScorer, SCORERS

classification_metrics = {"accuracy", "roc_auc", "average_precision", "neg_log_loss"}
for metric in ["precision", "recall", "f1"]:
    for average in ["macro", "micro", "samples", "weighted"]:
        classification_metrics.add(f"{metric}_{average}")

regression_metrics = {
    "explained_variance",
    "r2",
    "neg_mean_absolute_error",
    "neg_mean_squared_log_error",
    "neg_median_absolute_error",
    "neg_mean_squared_error",
}

all_metrics = {*classification_metrics, *regression_metrics}
reversed_scorers = {v: k for k, v in SCORERS.items()}


class MetricType(Enum):
    """ Metric types supported by GAMA. """

    CLASSIFICATION: int = 1  #: discrete target
    REGRESSION: int = 2  #: continuous target


class Metric:
    """ A thin layer around the `scorer` class of scikit-learn. """

    def __init__(self, scorer: Union[_BaseScorer, str]):
        if isinstance(scorer, str):
            scorer = get_scorer(scorer)
        if not isinstance(scorer, _BaseScorer):
            raise ValueError(
                "Scorer was not a valid scorer or could not be converted to one."
            )
        self.scorer = scorer
        self.name = reversed_scorers[scorer]
        self.requires_probabilities = (
            isinstance(scorer, _ProbaScorer) or self.name == "roc_auc"
        )

        if self.name in classification_metrics:
            self.task_type = MetricType.CLASSIFICATION
        elif self.name in regression_metrics:
            self.task_type = MetricType.REGRESSION
        else:
            raise ValueError(
                "Not sure which type of metric this is. Please raise an issue."
            )

        self.score = self.scorer._score_func

    def __call__(self, *args, **kwargs):
        return self.scorer(*args, **kwargs)

    def maximizable_score(self, *args, **kwargs):
        return self.scorer._sign * self.score(*args, **kwargs)


def scoring_to_metric(
    scoring: Union[str, Metric, Iterable[str], Iterable[Metric]]
) -> Tuple[Metric, ...]:
    if isinstance(scoring, str):
        return tuple([Metric(scoring)])
    if isinstance(scoring, Metric):
        return tuple([scoring])

    if isinstance(scoring, Iterable):
        if all([isinstance(scorer, (Metric, str)) for scorer in scoring]):
            converted_metrics = [
                scorer if isinstance(scorer, Metric) else Metric(scorer)
                for scorer in scoring
            ]
            return tuple(converted_metrics)

    raise TypeError("scoring must be str, Metric or Iterable (of str or Metric).")



from typing import Dict

from dash import Dash
from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import dash_bootstrap_components as dbc


def automark_slider(app: Dash, id_: str, label: str, slider_kwargs: Dict):
    defaults = dict(min=1, max=10, value=1, updatemode="drag")
    defaults.update(slider_kwargs)
    marks = {defaults["min"]: defaults["min"], defaults["max"]: defaults["max"]}

    cpu_input = dbc.FormGroup(
        [
            dbc.Label(label, html_for=id_, width=6),
            dbc.Col(dcc.Slider(id=id_, marks=marks, **defaults)),
        ],
        row=True,
    )
    app.callback.append(
        Output(id_, "marks"),
        [Input(id_, "value")],
        [State(id_, "min"), State(id_, "max")],
    )(_update_marks)
    return cpu_input


def _update_marks(selected_value, min_, max_):
    return {min_: str(min_), selected_value: str(selected_value), max_: str(max_)}


class ToggleButton:
    def __init__(self, button_id: str, app: Dash, label: str, start_on: bool = True):
        self._button_id = button_id
        self.html = self._build_content(label, start_on)

    def _build_content(self, label: str, start_on: bool):
        return dbc.FormGroup(
            [
                dbc.Label(label, html_for=self._button_id, width=6),
                dbc.Col(
                    dbc.Checklist(
                        id=self._button_id,
                        options=[{"label": "", "value": "on"}],
                        switch=True,
                        value="on" if start_on else "off",
                    )
                ),
            ],
            row=True,
        )



from typing import List, Union
from .terminal import DATA_TERMINAL, Terminal
from .primitive import Primitive


class PrimitiveNode:
    """ An instantiation for a Primitive with specific Terminals.

    Parameters
    ----------
    primitive: Primitive
        The Primitive type of this PrimitiveNode.
    data_node: PrimitiveNode
        The PrimitiveNode that specifies all preprocessing before this PrimitiveNode.
    terminals: List[Terminal]
        A list of terminals matching the `primitive`.
    """

    def __init__(
        self,
        primitive: Primitive,
        data_node: Union["PrimitiveNode", str],
        terminals: List[Terminal],
    ):
        self._primitive = primitive
        self._data_node = data_node
        self._terminals = sorted(terminals, key=lambda t: str(t))

    def __str__(self):
        """ Recursively stringify all primitive nodes (primitive and hyperparameters).

        Examples: - "GaussianNB(data)"
                  - "BernoulliNB(data, alpha=1.0)"
                  - "BernoulliNB(FastICA(data, tol=0.5), alpha=1.0)"
        """
        if self._terminals:
            terminal_str = ", ".join([repr(terminal) for terminal in self._terminals])
            return f"{self._primitive}({self._data_node}, {terminal_str})"
        else:
            return f"{self._primitive}({self._data_node})"

    @property
    def str_nonrecursive(self) -> str:
        """ Stringify all primitive node without data node (primitive and hyperparameters).

        Examples: - "GaussianNB()"
                  - "BernoulliNB(alpha=1.0)"
        """
        terminal_str = ", ".join([str(terminal) for terminal in self._terminals])
        return f"{self._primitive}({terminal_str})"

    def copy(self):
        """ Copies the object. Shallow for terminals, deep for data_node. """
        if self._data_node == DATA_TERMINAL:
            data_node_copy = DATA_TERMINAL
        else:
            data_node_copy = self._data_node.copy()
        return PrimitiveNode(
            primitive=self._primitive,
            data_node=data_node_copy,
            terminals=self._terminals.copy(),
        )

    @classmethod
    def from_string(cls, string: str, primitive_set: dict):
        """ Create a PrimitiveNode from string formatted like PrimitiveNode.__str__

        Parameters
        ----------
        string: str
            A string formatted similar to PrimitiveNode.__str__
        primitive_set: dict
            The dictionary defining all Terminals and Primitives.

        Returns
        -------
        PrimitiveNode
            The PrimitiveNode as defined the string.
        """
        # General form is
        # A(B(C(data[, C.param=value, ..])[, B.param=value, ..])[, A.param=value, ..])
        # below assumes that left parenthesis is never part of a parameter name or value
        primitives = string.split("(")[:-1]
        terminal_start_index = string.index(DATA_TERMINAL)
        terminals_string = string[terminal_start_index + len(DATA_TERMINAL) :]
        terminal_sets = terminals_string.split(")")[:-1]

        last_node: Union[PrimitiveNode, str] = DATA_TERMINAL
        for primitive_string, terminal_set in zip(reversed(primitives), terminal_sets):
            primitive = find_primitive(primitive_set, primitive_string)
            if terminal_set == "":
                terminals = []
            else:
                terminal_set = terminal_set[2:]  # 2 is because string starts with ', '
                terminals = [
                    find_terminal(primitive_set, terminal_string)
                    for terminal_string in terminal_set.split(", ")
                ]
            missing = set(primitive.input) - set(map(lambda t: t.identifier, terminals))
            if missing:
                raise ValueError(f"terminals {missing} for primitive {primitive}")
            last_node = cls(primitive, last_node, terminals)

        return last_node


def find_primitive(primitive_set: dict, primitive_string: str) -> Primitive:
    """ Find the Primitive that matches `primitive_string` in `primitive_set`. """
    all_primitives = primitive_set[DATA_TERMINAL] + primitive_set["prediction"]
    for primitive in all_primitives:
        if repr(primitive) == primitive_string:
            return primitive
    raise IndexError(f"Could not find Primitive of type '{primitive_string}'.")


def find_terminal(primitive_set: dict, terminal_string: str) -> Terminal:
    """ Find the Terminal that matches `terminal_string` in `primitive_set`. """
    term_type, _ = terminal_string.split("=")
    for terminal in primitive_set[term_type]:
        if repr(terminal) == terminal_string:
            return terminal
    raise RuntimeError(f"Could not find Terminal of type '{terminal_string}'.")



import numpy as np

from sklearn.naive_bayes import GaussianNB, BernoulliNB, MultinomialNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import (
    ExtraTreesClassifier,
    RandomForestClassifier,
    GradientBoostingClassifier,
)
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC
from sklearn.linear_model import LogisticRegression
from sklearn.cluster import FeatureAgglomeration
from sklearn.preprocessing import (
    MaxAbsScaler,
    MinMaxScaler,
    Normalizer,
    PolynomialFeatures,
    RobustScaler,
    StandardScaler,
    Binarizer,
)
from sklearn.kernel_approximation import Nystroem, RBFSampler
from sklearn.decomposition import PCA, FastICA
from sklearn.feature_selection import (
    SelectFwe,
    SelectPercentile,
    f_classif,
    VarianceThreshold,
)

# For comparison, this selection of operators and hyperparameters is
# currently most of what TPOT supports.

clf_config = {
    "alpha": [1e-3, 1e-2, 1e-1, 1.0, 10.0, 100.0],
    "fit_prior": [True, False],
    "min_samples_split": range(2, 21),
    "min_samples_leaf": range(1, 21),
    # Classifiers
    GaussianNB: {},
    BernoulliNB: {"alpha": [], "fit_prior": []},
    MultinomialNB: {"alpha": [], "fit_prior": []},
    DecisionTreeClassifier: {
        "criterion": ["gini", "entropy"],
        "max_depth": range(1, 11),
        "min_samples_split": [],
        "min_samples_leaf": [],
    },
    ExtraTreesClassifier: {
        "n_estimators": [100],
        "criterion": ["gini", "entropy"],
        "max_features": np.arange(0.05, 1.01, 0.05),
        "min_samples_split": [],
        "min_samples_leaf": [],
        "bootstrap": [True, False],
    },
    RandomForestClassifier: {
        "n_estimators": [100],
        "criterion": ["gini", "entropy"],
        "max_features": np.arange(0.05, 1.01, 0.05),
        "min_samples_split": range(2, 21),
        "min_samples_leaf": range(1, 21),
        "bootstrap": [True, False],
    },
    GradientBoostingClassifier: {
        "n_estimators": [100],
        "learning_rate": [1e-3, 1e-2, 1e-1, 0.5, 1.0],
        "max_depth": range(1, 11),
        "min_samples_split": range(2, 21),
        "min_samples_leaf": range(1, 21),
        "subsample": np.arange(0.05, 1.01, 0.05),
        "max_features": np.arange(0.05, 1.01, 0.05),
    },
    KNeighborsClassifier: {
        "n_neighbors": range(1, 51),
        "weights": ["uniform", "distance"],
        "p": [1, 2],
    },
    LinearSVC: {
        "penalty": ["l1", "l2"],
        "loss": ["hinge", "squared_hinge"],
        "dual": [False, True],
        "tol": [1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
        "C": [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],
        "param_check": [
            lambda params: (not params["dual"] or params["penalty"] == "l2")
            and not (params["penalty"] == "l1" and params["loss"] == "hinge")
            and not (
                params["penalty"] == "l2"
                and params["loss"] == "hinge"
                and not params["dual"]
            )
        ],
    },
    LogisticRegression: {
        "penalty": ["l2"],
        "C": [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1.0, 5.0, 10.0, 15.0, 20.0, 25.0],
        "dual": [False, True],
        "solver": ["lbfgs"],
    },
    Binarizer: {"threshold": np.arange(0.0, 1.01, 0.05)},
    FastICA: {"tol": np.arange(0.0, 1.01, 0.05)},
    FeatureAgglomeration: {
        "linkage": ["ward", "complete", "average"],
        "affinity": ["euclidean", "l1", "l2", "manhattan", "cosine", "precomputed"],
        "param_check": [
            lambda params: (not params["linkage"] == "ward")
            or params["affinity"] == "euclidean"
        ],
    },
    MaxAbsScaler: {},
    MinMaxScaler: {},
    Normalizer: {"norm": ["l1", "l2", "max"]},
    Nystroem: {
        "kernel": [
            "rbf",
            "cosine",
            "chi2",
            "laplacian",
            "polynomial",
            "poly",
            "linear",
            "additive_chi2",
            "sigmoid",
        ],
        "gamma": np.arange(0.0, 1.01, 0.05),
        "n_components": range(1, 11),
    },
    PCA: {"svd_solver": ["randomized"], "iterated_power": range(1, 11)},
    PolynomialFeatures: {
        "degree": [2],
        "include_bias": [False],
        "interaction_only": [False],
    },
    RBFSampler: {"gamma": np.arange(0.0, 1.01, 0.05)},
    RobustScaler: {},
    StandardScaler: {},
    # Selectors
    SelectFwe: {"alpha": np.arange(0, 0.05, 0.001), "score_func": {f_classif: None}},
    SelectPercentile: {"percentile": range(1, 100), "score_func": {f_classif: None}},
    VarianceThreshold: {"threshold": np.arange(0.05, 1.01, 0.05)},
}



""" A file selector component.
If tkinter is installed, this component is a button which opens the file dialog.
If tkinter is not installed, this component is a text input.
"""
import os
from typing import Optional, List

import dash_core_components as dcc
import dash_bootstrap_components as dbc
from dash.dependencies import Input

try:
    import tkinter
    from tkinter import filedialog

    tkinter_imported = True
except ModuleNotFoundError:
    tkinter_imported = False


class FileSelector:
    def __init__(
        self, id_: str, select_type: str = "file", filters: Optional[List[str]] = None
    ):
        """

        Parameters
        ----------
        id_: str
            id of the html element
        select_type: str, "file" or "folder" (default = "file")
            Whether the component should allow selection of a folder or a file.
        filters: List[str], optional (default = None)
            If set, a filter of allowed file type extensions.
        """
        self._id = id_
        self.select_type = select_type
        self.filters = filters
        self._listeners = []

    @property
    def html(self):
        if tkinter_imported:
            if self.select_type == "file":
                text = "#### Select File"
            else:
                text = "#### Select Folder"

            component = dbc.Button(
                [dcc.Markdown(text)], id=self._id, block=True, color="Link",
            )
        else:
            if self.select_type == "file":
                text = "Path to data file, e.g. ~/data/mydata.arff"
            else:
                text = "Path to folder, e.g. ~/data/"
            component = dbc.Input(id=self._id, placeholder=text, type="text",)
        return component

    def callback(self):
        if tkinter_imported:

            def get_file(_):
                tkinter.Tk().withdraw()
                if self.select_type == "file":
                    res = tkinter.filedialog.askopenfilename(filetypes=self.filters)
                else:
                    res = tkinter.filedialog.askdirectory()
                self._on_file_selected(res)

            field = "n_clicks"
        else:

            def get_file(text):
                if os.path.exists(text):
                    self._on_file_selected(text)

            field = "value"

        io_ = ([], [Input(self._id, field)])
        return io_, get_file

    def _on_file_selected(self, file: str):
        for listener in self._listeners:
            listener(file)



import datetime
import heapq
import logging
from typing import Tuple, List, Optional, Union, Dict
import pickle
import os

import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedShuffleSplit

from gama.genetic_programming.components import Individual

log = logging.getLogger(__name__)


class Evaluation:
    """ Record relevant evaluation data of an individual. """

    def __init__(
        self,
        individual: Individual,
        predictions: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]] = None,
        score: Tuple[float, ...] = (),
        estimators: Optional[List] = None,
        start_time: Optional[datetime.datetime] = None,
        duration: float = -1,
        error: str = None,
        pid: Optional[int] = None,
    ):
        self.individual: Individual = individual
        self.score = score
        self._estimators: Optional[List] = [] if estimators is None else estimators
        self.start_time = start_time
        self.duration = duration
        self.error = error
        self.pid = pid
        self._cache_file = None

        if isinstance(predictions, (pd.Series, pd.DataFrame)):
            predictions = predictions.values
        self._predictions: Optional[np.ndarray] = predictions

    def to_disk(self, directory):
        self._cache_file = os.path.join(directory, str(self.individual._id) + ".pkl")
        with open(self._cache_file, "wb") as fh:
            pickle.dump((self._estimators, self._predictions), fh)
        self._estimators, self._predictions = [], None

    def remove_from_disk(self):
        os.remove(os.path.join(self._cache_file))
        self._cache_file = None

    @property
    def estimators(self):
        if self._estimators or not self._cache_file:
            return self._estimators
        else:
            with open(self._cache_file, "rb") as fh:
                estimators, _ = pickle.load(fh)
                return estimators

    @property
    def predictions(self):
        if self._predictions is not None or not self._cache_file:
            return self._predictions
        else:
            with open(self._cache_file, "rb") as fh:
                _, predictions = pickle.load(fh)
                return predictions

    # Is there a better way to do this?
    # Assignment in __init__ is not preferred even if it saves lines.
    def __lt__(self, other):
        return self.score.__lt__(other.score)

    def __le__(self, other):
        return self.score.__le__(other.score)

    def __eq__(self, other):
        return self.score.__eq__(other.score)

    def __ne__(self, other):
        return self.score.__ne__(other.score)

    def __gt__(self, other):
        return self.score.__gt__(other.score)

    def __ge__(self, other):
        return self.score.__ge__(other.score)


class EvaluationLibrary:
    """ Maintains an in-memory record of evaluations.

    The main function of the EvaluationLibrary is to maintain a fast lookup for
    the best evaluations, and to discard meta-data of Evaluations which are not
    good enough.

    Specifically for the top `m` evaluations the estimators and predictions are kept.
    All but `n` predictions of each evaluation are discarded.
    As soon as an evaluation is no longer in the top `m`,
    its estimators and the sampled predictions are also discarded.
    Other evaluation meta-data (e.g. scores, evaluation time, errors) is not discarded.

    This discarding is useful to reduce memory usage when you know which meta-data
    is used later. E.g.:
    Ensembling selects from the best `m` models and uses `n` points for hill-climbing,
    but general score information is relevant for all evaluated models.
    """

    def __init__(
        self,
        m: Optional[int] = 200,
        n: Optional[int] = 10_000,
        sample: Optional[np.ndarray] = None,
        cache: str = "cache",
    ):
        """ Create an EvaluationLibrary for in-memory record of evaluations.

        Parameters
        ----------
        m: int, optional (default=200)
            Evaluations outside of the top `m` have their predictions and estimators
            discarded to reduce memory usage.
            The best `m` evaluations can be queried quickly through `n_best(int)`.
            If `None`, never discard predictions and estimators.
        n: int, optional (default=10_000)
            Instead of storing all predictions of the 'top m' evaluations,
            store only 'n' predictions. If left `None` all predictions will be kept.
            This parameter is ignored when 'sample' is also provided.
            Call `determine_sample_indices` before saving the first evaluation,
            if you want to ensure a class stratified sample.
        sample: np.ndarray, optional (default=None)
            Instead of storing all predictions of the 'top m' evaluations,
            store only those with indices specified in this array.
        cache: str (default="cache")
            Directory to save evaluations to.
            For large datasets or a big number of models,
            this can lead to memory issues.
        """
        self.top_evaluations: List[Evaluation] = []
        self.other_evaluations: List[Evaluation] = []
        self._m = m
        self._sample_n = n
        self.lookup: Dict[str, Evaluation] = {}
        self._cache = os.path.expandvars(cache)
        if not os.path.exists(self._cache):
            os.mkdir(self._cache)

        def main_node_str(e: Evaluation):
            return str(e.individual.main_node)

        self._lookup_key = main_node_str

        if sample is not None:
            self._sample = sample
        elif n is None:
            self._sample = None
        else:
            self._sample = "not set"

    @property
    def evaluations(self) -> List[Evaluation]:
        return self.top_evaluations + self.other_evaluations

    def determine_sample_indices(
        self,
        n: Optional[int] = None,
        prediction_size: Optional[int] = None,
        stratify: Optional[Union[np.ndarray, pd.Series, pd.DataFrame]] = None,
    ) -> None:
        """ Set `self._sample` to an array for sampling predictions or `None`.

        The sample indices can be class stratified if `stratify` is set.
        If `prediction_size` or `len(stratify)` is smaller than `n`,
        predictions will not be sampled and all predictions are saved.

        Parameters
        ----------
        n: int, optional (default=None)
            Number of predictions to keep for each evaluation.
            If `None`, use `n` set at the library's initialisation.
        prediction_size: int, optional (default=None)
            Number of predictions one can sample from.
        stratify: np.ndarray, pd.Series or pd.DataFrame, optional (default=None)
            The target variable of a supervised classification problem.
            The sample will be class stratified according to this variable.
        """
        if not ((prediction_size is None) ^ (stratify is None)):
            raise ValueError(
                "Exactly one of `prediction_size` and `stratify` must be set."
            )
        if len(self.top_evaluations) > 0:
            log.warning("New subsample not used for already stored evaluations.")
        n = self._sample_n if n is None else n

        if n is not None:
            if prediction_size is not None and n < prediction_size:
                # Subsample is to be chosen uniformly random.
                self._sample = np.random.choice(
                    range(prediction_size), size=n, replace=False
                )
            elif stratify is not None and n < len(stratify):
                splitter = StratifiedShuffleSplit(n_splits=1, train_size=n)
                self._sample, _ = next(
                    splitter.split(np.zeros(len(stratify)), stratify)
                )
            else:
                # Specified sample size exceeds size of predictions
                self._sample = None
        else:
            # No n was provided here nor set on initialization
            self._sample = None

    def _process_predictions(self, evaluation: Evaluation):
        """ Downsample evaluation predictions if required. """
        if self._sample_n == 0:
            evaluation._predictions = None
        if evaluation.predictions is None:
            return  # Predictions either not provided or removed because sample_n is 0.

        if isinstance(self._sample, str) and self._sample == "not set":
            # Happens only for the first evaluation with predictions.
            self.determine_sample_indices(self._sample_n, len(evaluation.predictions))

        if self._sample is not None:
            evaluation._predictions = evaluation.predictions[self._sample]

    def save_evaluation(self, evaluation: Evaluation) -> None:
        self._process_predictions(evaluation)

        if evaluation.error is not None:
            evaluation._estimators, evaluation._predictions = None, None
            self.other_evaluations.append(evaluation)
        elif self._m is None or self._m > len(self.top_evaluations):
            evaluation.to_disk(self._cache)
            heapq.heappush(self.top_evaluations, evaluation)
        else:
            removed = heapq.heappushpop(self.top_evaluations, evaluation)
            if removed == evaluation:
                # new evaluation is not in heap, big memory items may be discarded
                removed._predictions, removed._estimators = None, None
            else:
                # new evaluation is now on the heap, remove old from disk
                evaluation.to_disk(self._cache)
                removed.remove_from_disk()

            self.other_evaluations.append(removed)

        self.lookup[self._lookup_key(evaluation)] = evaluation

    def clear_cache(self):
        for file in os.listdir(self._cache):
            os.remove(os.path.join(self._cache, file))
        os.rmdir(self._cache)

    def n_best(self, n: int = 5, with_pipelines=True) -> List[Evaluation]:
        """ Return the best `n` pipelines.

        Slower if `n` exceeds `m` given on initialization.
        """
        if self._m is None or n <= self._m or with_pipelines:
            return heapq.nlargest(n, self.top_evaluations)
        else:
            return list(reversed(sorted(self.evaluations)))[:n]



from functools import partial
import logging
import math
from typing import List, Optional, Dict, Tuple, Any

import pandas as pd
import stopit

from gama.genetic_programming.operator_set import OperatorSet
from gama.logging.evaluation_logger import EvaluationLogger
from gama.search_methods.base_search import BaseSearch
from gama.utilities.generic.async_evaluator import AsyncEvaluator
from gama.genetic_programming.components.individual import Individual

log = logging.getLogger(__name__)


class AsynchronousSuccessiveHalving(BaseSearch):
    """ Asynchronous Halving Algorithm by Li et al.

    paper: https://arxiv.org/abs/1810.05934

    Parameters
    ----------
    reduction_factor: int, optional (default=3)
        Reduction factor of candidates between each rung.
    minimum_resource: int, optional (default=100)
        Number of samples to use in the lowest rung.
    maximum_resource: int, optional (default=number of samples in the dataset)
        Number of samples to use in the top rung.
        This should not exceed the number of samples in the data.
    minimum_early_stopping_rate: int (default=1)
        Number of lowest rungs to skip.
    """

    def __init__(
        self,
        reduction_factor: Optional[int] = None,
        minimum_resource: Optional[int] = None,
        maximum_resource: Optional[int] = None,
        minimum_early_stopping_rate: Optional[int] = None,
    ):
        super().__init__()
        # maps hyperparameter -> (set value, default)
        self._hyperparameters: Dict[str, Tuple[Any, Any]] = dict(
            reduction_factor=(reduction_factor, 3),
            minimum_resource=(minimum_resource, 100),
            maximum_resource=(maximum_resource, 100_000),
            minimum_early_stopping_rate=(minimum_early_stopping_rate, 1),
        )
        self.output = []

        self.logger = partial(
            EvaluationLogger,
            extra_fields=dict(rung=lambda e: e.individual.meta.get("rung", "unknown")),
        )

    def dynamic_defaults(self, x: pd.DataFrame, y: pd.DataFrame, time_limit: float):
        # `maximum_resource` is the number of samples used in the highest rung.
        # this typically should be the number of samples in the (training) dataset.
        self._overwrite_hyperparameter_default("maximum_resource", len(y))

    def search(self, operations: OperatorSet, start_candidates: List[Individual]):
        self.output = asha(
            operations, start_candidates=start_candidates, **self.hyperparameters
        )


def asha(
    operations: OperatorSet,
    start_candidates: List[Individual],
    reduction_factor: int = 3,
    minimum_resource: int = 100,
    maximum_resource: int = 100_000,
    minimum_early_stopping_rate: int = 1,
    max_full_evaluations: Optional[int] = None,
) -> List[Individual]:
    """ Asynchronous Halving Algorithm by Li et al.

    paper: https://arxiv.org/abs/1810.05934

    Parameters
    ----------
    operations: OperatorSet
        An operator set with `evaluate` and `individual` functions.
    start_candidates: List[Individual]
        A list which contains the set of best found individuals during search.
    reduction_factor: int (default=3)
        Reduction factor of candidates between each rung.
    minimum_resource: int (default=100)
        Number of samples to use in the lowest rung.
    maximum_resource: int (default=100_000)
        Number of samples to use in the top rung.
        This should not exceed the number of samples in the data.
    minimum_early_stopping_rate: int (default=1)
        Number of lowest rungs to skip.
    max_full_evaluations: Optional[int] (default=None)
        Maximum number of individuals to evaluate on the max rung (i.e. on all data).
        If None, the algorithm will be run indefinitely.

    Returns
    -------
    List[Individual]
        Individuals of the highest rung in which
        at least one individual has been evaluated.
    """

    # Note that here we index the rungs by all possible rungs (0..ceil(log_eta(R/r))),
    # and ignore the first minimum_early_stopping_rate rungs.
    # This contrasts the paper where rung 0 refers to the first used one.
    max_rung = math.ceil(
        math.log(maximum_resource / minimum_resource, reduction_factor)
    )
    rungs = range(minimum_early_stopping_rate, max_rung + 1)
    rung_resources = {
        rung: min(minimum_resource * (reduction_factor ** rung), maximum_resource)
        for rung in rungs
    }
    evaluate = partial(
        evaluate_on_rung, evaluate_individual=operations.evaluate, max_rung=max_rung
    )

    # Highest rungs first is how we typically iterate them
    # Should we just use lists of lists/heaps instead?
    rung_individuals: Dict[int, List[Tuple[float, Individual]]] = {
        rung: [] for rung in reversed(rungs)
    }
    promoted_individuals: Dict[int, List[Individual]] = {
        rung: [] for rung in reversed(rungs)
    }

    def get_job():
        for rung, individuals in list(rung_individuals.items())[1:]:
            # This is not in the paper code but is derived from fig 2b
            n_to_promote = math.floor(len(individuals) / reduction_factor)
            if n_to_promote - len(promoted_individuals[rung]) > 0:
                # Problem: equal loss falls back on comparison of individual
                not_promoted = set(individuals) - set(promoted_individuals[rung])
                if len(not_promoted) > 0:
                    to_promote = max(not_promoted, key=lambda i: i[0])
                    promoted_individuals[rung].append(to_promote)
                    return to_promote[1], rung + 1

        if start_candidates is not None and len(start_candidates) > 0:
            return start_candidates.pop(), minimum_early_stopping_rate
        else:
            return operations.individual(), minimum_early_stopping_rate

    try:
        with AsyncEvaluator() as async_:
            log.info("ASHA start")

            def start_new_job():
                individual, rung = get_job()
                time_penalty = rung_resources[rung] / max(rung_resources.values())
                async_.submit(
                    evaluate,
                    individual,
                    rung,
                    subsample=rung_resources[rung],
                    timeout=(10 + (time_penalty * 600)),
                )

            for _ in range(8):
                start_new_job()

            while (max_full_evaluations is None) or (
                len(rung_individuals[max_rung]) < max_full_evaluations
            ):
                future = operations.wait_next(async_)
                if future.result is not None:
                    rung = future.result.individual.meta["rung"]
                    loss = future.result.score[0]
                    individual = future.result.individual
                    rung_individuals[rung].append((loss, individual))
                start_new_job()

            highest_rung_reached = max(rungs)
    except stopit.TimeoutException:
        log.info("ASHA ended due to timeout.")
        reached_rungs = (rung for rung, inds in rung_individuals.items() if inds != [])
        highest_rung_reached = max(reached_rungs)
        if highest_rung_reached != max(rungs):
            raise RuntimeWarning("Highest rung not reached.")
    finally:
        for rung, individuals in rung_individuals.items():
            log.info(f"[{len(individuals)}] {rung}")
        return list(map(lambda p: p[1], rung_individuals[highest_rung_reached]))


def evaluate_on_rung(individual, rung, max_rung, evaluate_individual, *args, **kwargs):
    evaluation = evaluate_individual(individual, *args, **kwargs)
    evaluation.individual.meta["rung"] = rung
    # We want to avoid saving evaluations that are not on the max rung to disk,
    # because we only want to use pipelines evaluated on the max rung after search.
    # We're working on a better way to relay this information, this is temporary.
    if evaluation.error is None and rung != max_rung:
        evaluation.error = "Not a full evaluation."
    return evaluation



import inspect
from typing import Union, Optional

import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder

from .gama import Gama
from gama.data_loading import X_y_from_file
from gama.configuration.classification import clf_config
from gama.utilities.metrics import scoring_to_metric


class GamaClassifier(Gama):
    """ Gama with adaptations for (multi-class) classification. """

    def __init__(self, config=None, scoring="neg_log_loss", *args, **kwargs):
        if not config:
            # Do this to avoid the whole dictionary being included in the documentation.
            config = clf_config

        self._metrics = scoring_to_metric(scoring)
        if any(metric.requires_probabilities for metric in self._metrics):
            # we don't want classifiers that do not have `predict_proba`,
            # because then we have to start doing one hot encodings of predictions etc.
            config = {
                alg: hp
                for (alg, hp) in config.items()
                if not (
                    inspect.isclass(alg)
                    and issubclass(alg, ClassifierMixin)
                    and not hasattr(alg(), "predict_proba")
                )
            }

        self._label_encoder = None
        super().__init__(*args, **kwargs, config=config, scoring=scoring)

    def _predict(self, x: pd.DataFrame):
        """ Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array with predictions of shape (N,) where N is len(X).
        """
        y = self.model.predict(x)  # type: ignore
        # Decode the predicted labels - necessary only if ensemble is not used.
        if y[0] not in self._label_encoder.classes_:
            y = self._label_encoder.inverse_transform(y)
        return y

    def _predict_proba(self, x: pd.DataFrame):
        """ Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is len(x),
             and K is the number of class labels found in `y` of `fit`.
        """
        return self.model.predict_proba(x)  # type: ignore

    def predict_proba(self, x: Union[pd.DataFrame, np.ndarray]):
        """ Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            Data with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is len(x),
             and K is the number of class labels found in `y` of `fit`.
        """
        x = self._prepare_for_prediction(x)
        return self._predict_proba(x)

    def predict_proba_from_file(
        self,
        arff_file_path: str,
        target_column: Optional[str] = None,
        encoding: Optional[str] = None,
    ):
        """ Predict the class probabilities for input in the arff_file.

        Parameters
        ----------
        arff_file_path: str
            An ARFF file with the same columns as the one that used in fit.
            Target column must be present in file, but its values are ignored.
        target_column: str, optional (default=None)
            Specifies which column the model should predict.
            If left None, the last column is taken to be the target.
        encoding: str, optional
            Encoding of the ARFF file.

        Returns
        -------
        numpy.ndarray
            Numpy array with class probabilities.
            The array is of shape (N, K) where N is len(X),
            and K is the number of class labels found in `y` of `fit`.
        """
        x, _ = X_y_from_file(arff_file_path, target_column, encoding)
        x = self._prepare_for_prediction(x)
        return self._predict_proba(x)

    def fit(self, x, y, *args, **kwargs):
        """ Should use base class documentation. """
        y_ = y.squeeze() if isinstance(y, pd.DataFrame) else y
        self._label_encoder = LabelEncoder().fit(y_)
        if any([isinstance(yi, str) for yi in y_]):
            # If target values are `str` we encode them or scikit-learn will complain.
            y = self._label_encoder.transform(y_)
        self._evaluation_library.determine_sample_indices(stratify=y)
        super().fit(x, y, *args, **kwargs)

    def _encode_labels(self, y):
        self._label_encoder = LabelEncoder().fit(y)
        return self._label_encoder.transform(y)



import uuid
import copy
import logging
import random
import time
from typing import Optional, List, TYPE_CHECKING, Dict, Tuple, Sequence

import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.preprocessing import OneHotEncoder

from gama.genetic_programming.components import Individual
from gama.postprocessing.base_post_processing import BasePostProcessing
from gama.utilities.evaluation_library import EvaluationLibrary, Evaluation
from gama.utilities.export import (
    imports_and_steps_for_individual,
    format_import,
    format_pipeline,
    transformers_to_str,
)
from gama.utilities.metrics import Metric, MetricType


if TYPE_CHECKING:
    from gama.gama import Gama

log = logging.getLogger(__name__)


class EnsemblePostProcessing(BasePostProcessing):
    def __init__(
        self,
        time_fraction: float = 0.3,
        ensemble_size: Optional[int] = 25,
        hillclimb_size: Optional[int] = 10_000,
        max_models: Optional[int] = 200,
    ):
        """ Ensemble construction per Caruana et al.

        Parameters
        ----------
        time_fraction: float (default=0.3)
            Fraction of total time reserved for Ensemble building.
        ensemble_size: int, optional (default=25)
            Total number of models in the ensemble.
            When a single model is chosen more than once, it will increase its weight
            in the ensemble and *does* count towards this maximum.
        hillclimb_size: int, optional (default=10_000)
            Number of predictions that are used to determine the ensemble score
            during hillclimbing. If `None`, use all.
        max_models: int, optional (default=200)
            Only consider the best `max_models` number of models. If `None`, use all.
            Consequently also sets the max number of unique models in the ensemble.
        """
        super().__init__(time_fraction)
        self._hyperparameters = dict(
            ensemble_size=(ensemble_size, 25),
            metric=(None, None),
            evaluation_library=(None, None),
            hillclimb_size=(hillclimb_size, 10_000),
            max_models=(max_models, 200),
        )
        self._ensemble: Optional[Ensemble] = None

    def dynamic_defaults(self, gama: "Gama"):
        self._overwrite_hyperparameter_default("metric", gama._metrics[0])
        self._overwrite_hyperparameter_default(
            "evaluation_library", gama._evaluation_library
        )

    def post_process(
        self, x: pd.DataFrame, y: pd.Series, timeout: float, selection: List[Individual]
    ) -> object:
        self._ensemble = build_fit_ensemble(
            x,
            y,
            self.hyperparameters["ensemble_size"],
            timeout,
            self.hyperparameters["metric"],
            self.hyperparameters["evaluation_library"],
        )
        return self._ensemble

    def to_code(
        self, preprocessing: Sequence[Tuple[str, TransformerMixin]] = None
    ) -> str:
        if isinstance(self._ensemble, EnsembleClassifier):
            voter = "VotingClassifier"
        elif isinstance(self._ensemble, EnsembleRegressor):
            voter = "VotingRegressor"
        else:
            raise RuntimeError(f"Can't export ensemble of type {type(self._ensemble)}.")

        imports = {
            f"from sklearn.ensemble import {voter}",
            "from sklearn.pipeline import Pipeline",
        }

        pipelines = []
        for i, (model, weight) in enumerate(self._ensemble._models.values()):
            ind_imports, steps = imports_and_steps_for_individual(model.individual)
            imports = imports.union(ind_imports)
            pipeline_name = f"pipeline_{i}"
            pipelines.append(format_pipeline(steps, name=pipeline_name))

        estimators = ",".join([f"('{i}', pipeline_{i})" for i in range(len(pipelines))])
        weights = [weight for _, weight in self._ensemble._models.values()]

        if isinstance(self._ensemble, EnsembleClassifier):
            if self._ensemble._metric.requires_probabilities:
                voting = ",'soft'"
            else:
                voting = ", 'hard'"
        else:
            voting = ""  # This parameter does not exist for VotingRegressor

        if preprocessing is not None:
            imports = imports.union({format_import(t) for _, t in preprocessing})

        script = (
            "\n".join(sorted(imports))
            + "\n\n"
            + "\n\n".join(pipelines)
            + "\n"
            + f"ensemble = {voter}([{estimators}]{voting},{weights})\n"
        )
        if preprocessing is not None:
            trans_strs = transformers_to_str([t for _, t in preprocessing])
            names = [name for name, _ in preprocessing]
            steps = list(zip(names, trans_strs))
            script += format_pipeline(steps + [("ensemble", "ensemble")])
        return script


class Ensemble(object):
    def __init__(
        self,
        metric,
        y: pd.DataFrame,
        evaluation_library: EvaluationLibrary = None,
        shrink_on_pickle=True,
        downsample_to: Optional[int] = 10_000,
        use_top_n_only: Optional[int] = 200,
    ):
        """
        Either model_library or model_library_directory must be specified.
        If model_library is specified, model_library_directory is ignored.

        Parameters
        ----------
        metric: string or Metric
            Metric to optimize the ensemble towards.
        y: pandas.DataFrame
            True labels for the predictions made by the models in the library.
        evaluation_library: `gama.utilities.evaluation_library.EvaluationLibrary`
            A list of models from which an ensemble can be built.
        shrink_on_pickle: bool (default=True)
            If True, remove memory-intensive attributes that are required before pickle.
            When unpickled, the model can be used to create predictions,
            but the ensemble can't be changed.
        """
        if isinstance(metric, str):
            metric = Metric(metric)
        elif not isinstance(metric, Metric):
            raise ValueError(
                "metric must be specified as string or `gama.ea.metrics.Metric`."
            )

        if evaluation_library is None:
            raise ValueError(
                "`evaluation_library` is None but must be EvaluationLibrary."
            )
        elif not isinstance(evaluation_library, EvaluationLibrary):
            raise TypeError(
                "`evaluation_library` must be of type "
                "gama.utilities.evaluation_library.EvaluationLibrary."
            )

        if not isinstance(y, (pd.Series, pd.DataFrame)):
            raise TypeError(f"y_true must be pd.DataFrame or pd.Series, is {type(y)}.")

        self._metric = metric
        self.evaluation_library = evaluation_library
        self._model_library: List[Evaluation] = []
        self._use_top_n_only = use_top_n_only
        self._shrink_on_pickle = shrink_on_pickle
        self._prediction_transformation = None

        if self.evaluation_library._sample is not None:
            # If the library stores sampled predictions, we match that first.
            y = y.iloc[self.evaluation_library._sample]

        # Then apply even more sampling if requested.
        if downsample_to is None or downsample_to >= len(y):
            if downsample_to is not None:
                log.info(f"Not downsampling because only {len(y)} samples were stored.")
            self._y = y
            self._prediction_sample = None
        else:
            log.info(f"Downsampling as training data exceeds {downsample_to} samples.")
            self._prediction_sample = random.sample(range(len(y)), downsample_to)
            self._y = y.iloc[self._prediction_sample]

        self._internal_score = -float("inf")
        self._fit_models = None
        self._maximize = True
        self._models: Dict[uuid.UUID, Tuple[Evaluation, int]] = {}

    @property
    def model_library(self):
        if not self._model_library:
            self._model_library = []
            for evaluation in self.evaluation_library.n_best(self._use_top_n_only):
                predictions = evaluation.predictions
                if self._prediction_transformation:
                    predictions = self._prediction_transformation(predictions)
                if self._prediction_sample:
                    predictions = predictions[self._prediction_sample]

                e = copy.copy(evaluation)
                e._predictions = predictions
                self._model_library.append(e)

        return self._model_library

    def _ensemble_validation_score(self, prediction_to_validate=None):
        raise NotImplementedError("Must be implemented by child class.")

    def _total_fit_weights(self):
        return sum([weight for (model, weight) in self._fit_models])

    def _total_model_weights(self):
        return sum([weight for (model, weight) in self._models.values()])

    def _averaged_validation_predictions(self):
        """ Weighted average of predictions of current models on the hillclimb set. """
        weighted_sum_predictions = sum(
            [model.predictions * weight for (model, weight) in self._models.values()]
        )
        return weighted_sum_predictions / self._total_model_weights()

    def build_initial_ensemble(self, n: int):
        """ Add top n models in EvaluationLibrary to the ensemble.

        Parameters
        ----------
        n: int
            Number of models to include.
        """
        if not n > 0:
            raise ValueError("Ensemble must include at least one model.")
        if self._models:
            log.warning(
                "The ensemble already contained models. Overwriting the ensemble."
            )
            self._models = {}

        # Since the model library only features unique models,
        # we do not need to check for duplicates here.
        for model in self.model_library[:n]:
            self._add_model(model)

        log.debug(
            "Initial ensemble created with score {}".format(
                self._ensemble_validation_score()
            )
        )

    def _add_model(self, model, add_weight=1):
        """ Add a specific model to the ensemble or increases its weight. """
        model, weight = self._models.pop(model.individual._id, (model, 0))
        new_weight = weight + add_weight
        self._models[model.individual._id] = (model, new_weight)
        log.debug(f"Weight {model.individual.short_name('>')} set to {new_weight}.")

    def expand_ensemble(self, n: int):
        """ Adds new models to the ensemble based on earlier given data.

        Parameters
        ----------
        n: int
            Number of models to add to current ensemble.
        """
        if not n > 0:
            raise ValueError("n must be greater than 0.")

        for _ in range(n):
            best_addition_score = -float("inf")
            current_weighted_average = self._averaged_validation_predictions()
            current_total_weight = self._total_model_weights()
            for model in self.model_library:
                if model.score == 0:
                    continue
                candidate_pred = current_weighted_average + (
                    model.predictions - current_weighted_average
                ) / (current_total_weight + 1)
                candidate_ensemble_score = self._ensemble_validation_score(
                    candidate_pred
                )
                if best_addition_score < candidate_ensemble_score:
                    best_addition, best_addition_score = model, candidate_ensemble_score

            self._add_model(best_addition)
            self._internal_score = best_addition_score
            log.info(
                "Ensemble size {} , best score: {}".format(
                    self._total_model_weights(), best_addition_score
                )
            )

    def fit(self, x, y, timeout=1e6):
        """ Constructs an Ensemble out of the library of models.

        Parameters
        ----------
        x:
            Data to fit the final selection of models on.
        y:
            Targets corresponding to features x.
        timeout: int (default=1e6)
            Maximum amount of time in seconds that is allowed for fitting pipelines.
            If this time is exceeded, only pipelines fit until that point are taken
            into account when making predictions.
            Starting the parallelization takes roughly 4 seconds by itself.
        """
        if not self._models:
            raise RuntimeError(
                "You need to call `build` to select models before fitting them."
            )
        if timeout <= 0:
            raise ValueError("timeout must be greater than 0.")

        self._fit_models = [
            (estimator, weight)
            for (model, weight) in self._models.values()
            for estimator in model.estimators
        ]
        # for (model, weight) in self._models.values():

        # self._fit_models = []
        # futures = set()
        # with stopit.ThreadingTimeout(timeout) as c_mgr, AsyncEvaluator() as async_:
        #     for (model, weight) in self._models.values():
        #         futures.add(
        #           async_.submit(fit_and_weight, (model.pipeline, X, y, weight))
        #           )
        #
        #     for _ in self._models.values():
        #         future = async_.wait_next()
        #         pipeline, weight = future.result
        #         if weight > 0:
        #             self._fit_models.append((pipeline, weight))
        #
        # if not c_mgr:
        #     log.info("Fitting of ensemble stopped early.")

    def _get_weighted_mean_predictions(self, X, predict_method="predict"):
        weighted_predictions = []
        for (model, weight) in self._fit_models:
            target_prediction = getattr(model, predict_method)(X)
            if self._prediction_transformation:
                target_prediction = self._prediction_transformation(target_prediction)
            weighted_predictions.append(target_prediction * weight)

        return sum(weighted_predictions) / self._total_fit_weights()

    def __str__(self):
        if not self._models:
            return "Ensemble with no models."
        _str_ = f"Ensemble of {len(self._models)} pipelines.\nR\tW\tScore\tPipeline\n"
        models = sorted(self._models.values(), key=lambda x: x[0].validation_score)
        for i, (model, weight) in enumerate(models):
            _str_ += f"{i}\t{weight}\t{model.validation_score[0]:.4f}\t{model.name}\n"
        return _str_

    def __getstate__(self):
        if self._shrink_on_pickle:
            log.info(
                "Shrinking before pickle because shrink_on_pickle is True."
                "Removing anything that is not needed for predict-functionality."
                "Functionality to expand ensemble after unpickle is not available."
            )
            self._models = None
            self._model_library = None
            # self._y_true can not be removed as it is needed to ensure proper
            # dimensionality of predictions.
            # Alternatively, one could just save the number of classes instead.

        return self.__dict__.copy()


def fit_and_weight(args):
    """ Fit the pipeline given the data. Update weight to 0 if fitting fails.

    Parameters
    ----------
    args: Tuple
        Expected Tuple of [Pipeline, X, y, weight].

    Returns
    -------
    pipeline
        The same pipeline that was provided as input.
    weight
        If fitting succeeded, return the input weight.
        If *any* exception occurred during fitting, weight is 0.
    """
    pipeline, X, y, weight = args
    try:
        pipeline.fit(X, y)
    except Exception:
        log.warning(f"Exception fitting {pipeline}. Set weight to 0.", exc_info=True)
        weight = 0

    return pipeline, weight


class EnsembleClassifier(Ensemble):
    def __init__(self, metric, y_true, label_encoder=None, *args, **kwargs):
        super().__init__(metric, y_true, *args, **kwargs)
        self._label_encoder = label_encoder

        # For metrics that only require class labels,
        # we still want to apply one-hot-encoding to average predictions.
        y_as_squeezed_array = y_true.to_numpy().reshape(-1, 1)
        self._one_hot_encoder = OneHotEncoder(categories="auto")
        self._one_hot_encoder.fit(y_as_squeezed_array)

        if self._metric.requires_probabilities:
            self._y = self._one_hot_encoder.transform(
                self._y.to_numpy().reshape(-1, 1)
            ).toarray()
            if self._prediction_sample is not None:
                self._y = self._y[self._prediction_sample]
        else:

            def one_hot_encode_predictions(predictions):
                return self._one_hot_encoder.transform(predictions.reshape(-1, 1))

            self._prediction_transformation = one_hot_encode_predictions

    def _ensemble_validation_score(self, prediction_to_validate=None):
        if prediction_to_validate is None:
            prediction_to_validate = self._averaged_validation_predictions()

        if self._metric.requires_probabilities:
            return self._metric.maximizable_score(self._y, prediction_to_validate)
        else:
            # argmax returns (N, 1) matrix, need to squeeze it to (N,) for scoring.
            class_predictions = self._one_hot_encoder.inverse_transform(
                prediction_to_validate.toarray()
            )
            return self._metric.maximizable_score(self._y, class_predictions)

    def predict(self, X):
        if self._metric.requires_probabilities:
            log.warning(
                "Ensemble was tuned with a class-probabilities metric. "
                "Using argmax of probabilities, which may not give optimal predictions."
            )
            class_probabilities = self._get_weighted_mean_predictions(
                X, "predict_proba"
            )
        else:
            class_probabilities = self._get_weighted_mean_predictions(
                X, "predict"
            ).toarray()

        class_predictions = self._one_hot_encoder.inverse_transform(class_probabilities)
        if self._label_encoder:
            class_predictions = self._label_encoder.inverse_transform(class_predictions)

        return class_predictions.ravel()

    def predict_proba(self, X):
        if self._metric.requires_probabilities:
            return self._get_weighted_mean_predictions(X, "predict_proba")
        else:
            log.warning(
                "Ensemble was tuned with a class label predictions metric, "
                "not probabilities. Using weighted mean of class predictions."
            )
            return self._get_weighted_mean_predictions(X, "predict").toarray()


class EnsembleRegressor(Ensemble):
    def _ensemble_validation_score(self, prediction_to_validate=None):
        if prediction_to_validate is None:
            prediction_to_validate = self._averaged_validation_predictions()
        return self._metric.maximizable_score(self._y, prediction_to_validate)

    def predict(self, X):
        return self._get_weighted_mean_predictions(X)


def build_fit_ensemble(
    x,
    y,
    ensemble_size: int,
    timeout: float,
    metric: Metric,
    evaluation_library: EvaluationLibrary,
    encoder: Optional[object] = None,
) -> Ensemble:
    """ Construct an Ensemble of models, optimizing for metric. """
    start_build = time.time()

    log.debug("Building ensemble.")
    if metric.task_type == MetricType.REGRESSION:
        ensemble = EnsembleRegressor(metric, y, evaluation_library)  # type: Ensemble
    elif metric.task_type == MetricType.CLASSIFICATION:
        ensemble = EnsembleClassifier(metric, y, evaluation_library=evaluation_library)
        ensemble._label_encoder = encoder
    else:
        raise ValueError(f"Unknown metric task type {metric.task_type}")

    try:
        # Starting with more models in the ensemble should help against overfitting,
        # but depending on the total ensemble size, it might leave too little room
        # to calibrate the weights or add new models. So we have some adaptive defaults.
        if ensemble_size <= 10:
            ensemble.build_initial_ensemble(1)
        else:
            ensemble.build_initial_ensemble(10)

        remainder = ensemble_size - ensemble._total_model_weights()
        if remainder > 0:
            ensemble.expand_ensemble(remainder)

        build_time = time.time() - start_build
        timeout = timeout - build_time
        log.info(f"Ensemble build took {build_time}s. Fit with timeout {timeout}s.")
        ensemble.fit(x, y, timeout)
    except Exception as e:
        log.warning(f"Error during auto ensemble: {e}", exc_info=True)

    return ensemble



import logging
from typing import Optional, Iterator, List, Tuple
import category_encoders as ce
import pandas as pd
from sklearn.base import TransformerMixin
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline

log = logging.getLogger(__name__)


def select_categorical_columns(
    df: pd.DataFrame,
    min_f: Optional[int] = None,
    max_f: Optional[int] = None,
    ignore_nan: bool = True,
) -> Iterator[str]:
    """ Find all categorical columns with at least `min_f` and at most `max_f` factors.

    Parameters
    ----------
    df: pd.DataFrame
        Pandas DataFrame to design the encoder for.
    min_f: int, optional (default=None)
        The inclusive minimum number of unique values the column should have.
    max_f: int, optional (default=None)
        The inclusive maximum number of unique values the column should have.
    ignore_nan: bool (default=True)
        If True, don't count NaN as a unique value.
        If False, count NaN as a unique value (only once).

    Returns
    -------
    An iterator which iterates over the column names that satisfy the criteria.
    """
    for column in df:
        if isinstance(df[column].dtype, pd.core.dtypes.dtypes.CategoricalDtype):
            nfactors = df[column].nunique(dropna=ignore_nan)
            if (min_f is None or min_f <= nfactors) and (
                max_f is None or nfactors <= max_f
            ):
                yield column


def basic_encoding(x: pd.DataFrame):
    """ Perform 'basic' encoding of categorical features.

     Specifically, perform:
      - Ordinal encoding for features with 2 or fewer unique values.
      - One hot encoding for features with at most 10 unique values.
     """
    binary_features = list(select_categorical_columns(x, max_f=2))
    leq_10_features = list(select_categorical_columns(x, min_f=3, max_f=10))

    encoding_pipeline = Pipeline(
        steps=[
            ("ord-enc", ce.OrdinalEncoder(cols=binary_features, drop_invariant=True)),
            ("oh-enc", ce.OneHotEncoder(cols=leq_10_features, handle_missing="ignore")),
        ]
    )
    x_enc = encoding_pipeline.fit_transform(x, y=None)  # Is this allowed?
    return x_enc, encoding_pipeline


def basic_pipeline_extension(x: pd.DataFrame) -> List[Tuple[str, TransformerMixin]]:
    """ Define a TargetEncoder and SimpleImputer.

    TargetEncoding is will encode categorical features with more than 10 unique values.
    SimpleImputer imputes with the median.
    """
    many_factor_features = list(select_categorical_columns(x, min_f=11))
    return [
        ("target_enc", ce.TargetEncoder(cols=many_factor_features)),
        ("imputation", SimpleImputer(strategy="median")),
    ]



"""
I don't want to be reinventing the wheel but I can't find a satisfying implementation.
I want to be able to execute arbitrary functions asynchronously on a different process.
Any ongoing subprocesses must immediately be able to be terminated without errors.
Results of cancelled subprocesses may be ignored.

`concurrent.futures.ProcessPoolExecutor` gets very close to the desired implementation,
but it has issues:
    - by default it waits for subprocesses to close on __exit__.
      Unfortunately it is possible the subprocesses can be running non-Python code,
      e.g. a C implementation of SVC whose subprocess won't end until fit is complete.
    - even if that is overwritten and no wait is performed,
      the subprocess will raise an error when it is done.
      Though that does not hinder the execution of the program,
      I don't want errors for expected behavior.
"""
import datetime
import gc
import logging
import multiprocessing
import os
import psutil
import queue
import struct
import time
import traceback
from typing import Optional, Callable, Dict, List
import uuid

from psutil import NoSuchProcess

try:
    import resource
except ModuleNotFoundError:
    resource = None  # type: ignore


log = logging.getLogger(__name__)


class AsyncFuture:
    """ Reference to a function call executed on a different process. """

    def __init__(self, fn, *args, **kwargs):
        self.id = uuid.uuid4()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.exception = None
        self.traceback = None

    def execute(self, extra_kwargs):
        """ Execute the function call `fn(*args, **kwargs)` and record results. """
        try:
            # Don't update self.kwargs, as it will be pickled back to the main process
            kwargs = {**self.kwargs, **extra_kwargs}
            self.result = self.fn(*self.args, **kwargs)
        except Exception as e:
            self.exception = e
            self.traceback = traceback.format_exc()


class AsyncEvaluator:
    """ Manages subprocesses on which arbitrary functions can be evaluated.

    The function and all its arguments must be picklable.
    Using the same AsyncEvaluator in two different contexts raises a `RuntimeError`.

    defaults: Dict, optional (default=None)
        Default parameter values shared between all submit calls.
        This allows these defaults to be transferred only once per process,
        instead of twice per call (to and from the subprocess).
        Only supports keyword arguments.
    """

    defaults: Dict = {}

    def __init__(
        self,
        n_workers: Optional[int] = None,
        memory_limit_mb: Optional[int] = None,
        logfile: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        n_workers : int, optional (default=None)
            Maximum number of subprocesses to run for parallel evaluations.
            Defaults to `AsyncEvaluator.n_jobs`, using all cores unless overwritten.
        memory_limit_mb : int, optional (default=None)
            The maximum number of megabytes that this process and its subprocesses
            may use in total. If None, no limit is enforced.
            There is no guarantee the limit is not violated.
        logfile : str, optional (default=None)
            If set, recorded resource usage will be written to this file.
        """
        self._has_entered = False
        self.futures: Dict[uuid.UUID, AsyncFuture] = {}
        self._processes: List[psutil.Process] = []
        self._n_jobs = n_workers
        self._memory_limit_mb = memory_limit_mb
        self._mem_violations = 0
        self._mem_behaved = 0
        self._logfile = logfile

        self._queue_manager = multiprocessing.Manager()
        self._input = self._queue_manager.Queue()
        self._output = self._queue_manager.Queue()
        pid = os.getpid()
        self._main_process = psutil.Process(pid)

    def __enter__(self):
        if self._has_entered:
            raise RuntimeError(
                "You can not use the same AsyncEvaluator in two different contexts."
            )
        self._has_entered = True

        self._input = self._queue_manager.Queue()
        self._output = self._queue_manager.Queue()

        log.debug(
            f"Process {self._main_process.pid} starting {self._n_jobs} subprocesses."
        )
        for _ in range(self._n_jobs):
            self._start_worker_process()
            # if resource and AsyncEvaluator.memory_limit_mb:
            #     limit = AsyncEvaluator.memory_limit_mb * (2 ** 20)
            #     resource.prlimit(subprocess.pid, resource.RLIMIT_AS, (limit, limit))

        self._log_memory_usage()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        log.debug(f"Terminating {len(self._processes)} subprocesses.")
        # This is ugly as the subprocesses use shared queues.
        # It is in direct conflict with guidelines:
        # https://docs.python.org/3/library/multiprocessing.html#all-start-methods
        for subprocess in self._processes:
            try:
                subprocess.terminate()
            except psutil.NoSuchProcess:
                log.debug(f"Daemon {subprocess.pid} stopped prematurely.")
        return False

    def submit(self, fn: Callable, *args, **kwargs) -> AsyncFuture:
        """ Submit fn(*args, **kwargs) to be evaluated on a subprocess.

        Parameters
        ----------
        fn: Callable
            Function to call on a subprocess.
        args
            Positional arguments to call `fn` with.
        kwargs
            Keyword arguments to call `fn` with.

        Returns
        -------
        AsyncFuture
            A Future of which the `result` or `exception` field will be populated
            once evaluation is finished.
        """
        future = AsyncFuture(fn, *args, **kwargs)
        self.futures[future.id] = future
        self._input.put(future)
        return future

    def wait_next(self, poll_time: float = 0.05) -> AsyncFuture:
        """ Wait until an AsyncFuture has been completed and return it.


        Parameters
        ----------
        poll_time: float (default=0.05)
            Time to sleep between checking if a future has been completed.

        Returns
        -------
        AsyncFuture
            The completed future that completed first.

        Raises
        ------
        RuntimeError
            If all futures have already been completed and returned.
        """
        if len(self.futures) == 0:
            raise RuntimeError("No Futures queued, must call `submit` first.")

        while True:
            try:
                self._control_memory_usage()
                self._log_memory_usage()
                completed_future = self._output.get(block=False)
                match = self.futures.pop(completed_future.id)
                match.result, match.exception, match.traceback = (
                    completed_future.result,
                    completed_future.exception,
                    completed_future.traceback,
                )
                self._mem_behaved += 1
                return match
            except queue.Empty:
                time.sleep(poll_time)
                continue
            except multiprocessing.managers.RemoteError as e:
                log.warning(f"Encountered RemoteError: {str(e)}", exc_info=True)
                time.sleep(poll_time)

    def _start_worker_process(self) -> psutil.Process:
        """ Start a new worker node and add it to the process pool. """
        mp_process = multiprocessing.Process(
            target=evaluator_daemon,
            args=(self._input, self._output, AsyncEvaluator.defaults),
            daemon=True,
        )
        mp_process.start()
        subprocess = psutil.Process(mp_process.pid)
        self._processes.append(subprocess)
        return subprocess

    def _stop_worker_process(self, process: psutil.Process):
        """ Terminate a new worker node and remove it from the process pool. """
        process.terminate()
        self._processes.remove(process)

    def _control_memory_usage(self, threshold=0.05):
        """ Dynamically restarts or kills processes to adhere to memory constraints. """
        if self._memory_limit_mb is None:
            return
        # If the memory usage of all processes (the main process, and the evaluation
        # subprocesses) exceeds the maximum allowed memory usage, we have to terminate
        # one of them.
        # If we were never to start new processes, eventually all subprocesses would
        # likely be killed due to 'silly' pipelines (e.g. multiple polynomial feature
        # steps).
        # On the other hand if there is e.g. a big dataset, by always restarting we
        # will set up the same scenario for failure over and over again.
        # So we want to dynamically find the right amount of evaluation processes, such
        # that the total memory usage is not exceeded "too often".
        # Here `threshold` defines the ratio of processes that should be allowed to
        # fail due to memory constraints. Setting it too high might lead to aggressive
        # subprocess killing and underutilizing compute resources. If it is too low,
        # the number of concurrent jobs might shrink too slowly inducing a lot of
        # loss in compute time due to interrupted evaluations.
        # ! Like the rest of this module, I hate to use custom code with this,
        # in particular there is a risk that terminating the process might leave
        # the multiprocess queue broken.
        mem_proc = list(self._get_memory_usage())
        if sum(map(lambda x: x[1], mem_proc)) > self._memory_limit_mb:
            log.info(
                f"GAMA exceeded memory usage "
                f"({self._mem_violations}, {self._mem_behaved})."
            )
            self._log_memory_usage()
            self._mem_violations += 1
            # Find the process with the most memory usage, that is not the main process
            proc, _ = max(mem_proc[1:], key=lambda t: t[1])
            n_evaluations = self._mem_violations + self._mem_behaved
            fail_ratio = self._mem_violations / n_evaluations
            if fail_ratio < threshold or len(self._processes) == 1:
                # restart `pid`
                log.info(f"Terminating {proc.pid} due to memory usage.")
                self._stop_worker_process(proc)
                log.info("Starting new evaluations process.")
                self._start_worker_process()
            else:
                # More than one process left alive and a violation of the threshold,
                # requires killing a subprocess.
                self._mem_behaved = 0
                self._mem_violations = 0
                log.info(f"Terminating {proc.pid} due to memory usage.")
                self._stop_worker_process(proc)
            # todo: update the Future of the evaluation that was terminated.

    def _log_memory_usage(self):
        if not self._logfile:
            return
        mem_by_pid = self._get_memory_usage()
        mem_str = ",".join([f"{proc.pid},{mem_mb}" for (proc, mem_mb) in mem_by_pid])
        timestamp = datetime.datetime.now().isoformat()

        with open(self._logfile, "a") as memory_log:
            memory_log.write(f"{timestamp},{mem_str}\n")

    def _get_memory_usage(self):
        processes = [self._main_process] + self._processes
        for process in processes:
            try:
                yield process, process.memory_info()[0] / (2 ** 20)
            except NoSuchProcess:
                # can never be the main process anyway
                self._processes = [p for p in self._processes if p.pid != process.pid]
                self._start_worker_process()


def evaluator_daemon(
    input_queue: queue.Queue,
    output_queue: queue.Queue,
    default_parameters: Optional[Dict] = None,
):
    """ Function for daemon subprocess that evaluates functions from AsyncFutures.

    Parameters
    ----------
    input_queue: queue.Queue[AsyncFuture]
        Queue to get AsyncFuture from.
        Queue should be managed by multiprocessing.manager.
    output_queue: queue.Queue[AsyncFuture]
        Queue to put AsyncFuture to.
        Queue should be managed by multiprocessing.manager.
    default_parameters: Dict, optional (default=None)
        Additional parameters to pass to AsyncFuture.Execute.
        This is useful to avoid passing lots of repetitive data through AsyncFuture.
    """
    try:
        while True:
            try:
                future = input_queue.get()
                future.execute(default_parameters)
                if future.result:
                    if isinstance(future.result, tuple):
                        result = future.result[0]
                    else:
                        result = future.result
                    if isinstance(result.error, MemoryError):
                        # Can't pickle MemoryErrors. Should work around this later.
                        result.error = "MemoryError"
                        gc.collect()
                output_queue.put(future)
            except (MemoryError, struct.error) as e:
                future.result = None
                future.exception = str(type(e))
                gc.collect()
                output_queue.put(future)
    except Exception as e:
        # There are no plans currently for recovering from any exception:
        print(f"Stopping daemon:{type(e)}:{str(e)}")
        traceback.print_exc()
    # Not sure if required: give main process time to process log message.
    time.sleep(5)



""" Functions which take two Individuals and produce at least one new Individual. """
import random
from typing import List, Callable, Iterable, Optional, Tuple

from gama.genetic_programming.components import Individual


def random_crossover(
    ind1: Individual, ind2: Individual, max_length: Optional[int] = None
) -> Tuple[Individual, Individual]:
    """ Random valid crossover between two individuals in-place, if it can be done.

    Parameters
    ----------
    ind1: Individual
        The individual to crossover with ind2.
    ind2: Individual
        The individual to crossover with ind1.
    max_length: int, optional(default=None)
        The first individual in the returned tuple has at most `max_length` primitives.
        Requires both provided individuals to contain at most `max_length` primitives.

    Raises
    ------
    ValueError
        - If there is no valid crossover function for the two individuals.
        - If `max_length` is set and either `ind1` or `ind2` contain
          more primitives than `max_length`.
    """
    if max_length is not None and len(ind1.primitives) > max_length:
        raise ValueError(f"`individual1` ({ind1}) exceeds `max_length` ({max_length}).")
    if max_length is not None and len(ind2.primitives) > max_length:
        raise ValueError(f"`individual2` ({ind2}) exceeds `max_length` ({max_length}).")

    crossover_choices = _valid_crossover_functions(ind1, ind2)
    if len(crossover_choices) == 0:
        raise ValueError(f"{ind1.pipeline_str()} and {ind2.pipeline_str()} can't mate.")
    ind1, ind2 = random.choice(crossover_choices)(ind1, ind2)

    if max_length is not None and len(ind1.primitives) > max_length:
        return ind2, ind1
    return ind1, ind2


def crossover_primitives(
    ind1: Individual, ind2: Individual
) -> Tuple[Individual, Individual]:
    """ Crossover two individuals by exchanging any number of preprocessing steps.

    Parameters
    ----------
    ind1: Individual
        The individual to crossover with individual2.
    ind2: Individual
        The individual to crossover with individual1.
    """
    p1_node = random.choice(list(ind1.primitives)[:-1])
    p2_node = random.choice(list(ind2.primitives)[:-1])
    p1_node._data_node, p2_node._data_node = p2_node._data_node, p1_node._data_node
    return ind1, ind2


def crossover_terminals(
    ind1: Individual, ind2: Individual
) -> Tuple[Individual, Individual]:
    """ Crossover two individuals in-place by exchanging two Terminals.

    Terminals must share output type but have different values.

    Parameters
    ----------
    ind1: Individual
        The individual to crossover with individual2.
    ind2: Individual
        The individual to crossover with individual1.
    """
    options = _shared_terminals(ind1, ind2, with_indices=True, value_match="different")
    i, ind1_term, j, ind2_term = random.choice(list(options))
    ind1.replace_terminal(i, ind2_term)
    ind2.replace_terminal(j, ind1_term)
    return ind1, ind2


def _shared_terminals(
    ind1: Individual,
    ind2: Individual,
    with_indices: bool = True,
    value_match: str = "different",
) -> Iterable:
    """ Finds all shared Terminals between two Individuals.

    Parameters
    ----------
    ind1: Individual
    ind2: Individual
    with_indices: bool (default=True)
        If True, also return the indices of the Terminals w.r.t. the Individual.
    value_match: str (default='different')
        Indicates with matches to return, based on terminal values.
        Accepted values are:

         - 'different': only return terminals with different values from each other
         - 'equal': only return terminals with equal values
         - 'all': return all shared terminals regardless of value

    Returns
    -------
    Sequence of Tuples with both Terminals, with the Terminal from ind1 first.
        Tuple[Terminal, Terminal] if `with_indices` is False
        Tuple[int, Terminal, int, Terminal] if `with_indices` is True,
        each int specifies the index of the Terminal directly after.
    """
    if value_match not in ["different", "equal", "all"]:
        raise ValueError(f"`value_match` ('{value_match}') is not valid.")

    for i, ind1_term in enumerate(ind1.terminals):
        for j, ind2_term in enumerate(ind2.terminals):
            if ind1_term.identifier == ind2_term.identifier:
                if value_match == "different" and ind1_term.value == ind2_term.value:
                    continue
                if value_match == "equal" and ind1_term.value != ind2_term.value:
                    continue
                if with_indices:
                    yield (i, ind1_term, j, ind2_term)
                else:
                    yield (ind1_term, ind2_term)


def _valid_crossover_functions(ind1: Individual, ind2: Individual) -> List[Callable]:
    """ Find all crossover functions that can produce new individuals from this input.

    Parameters
    ----------
    ind1: Individual
        The individual to crossover with individual2.
    ind2: Individual
        The individual to crossover with individual1.

    Returns
    -------
    List[Callable]
        List of valid crossover functions given the content of ind1 and ind2.
    """
    crossover_choices = []
    if list(_shared_terminals(ind1, ind2)):
        crossover_choices.append(crossover_terminals)
    if len(list(ind1.primitives)) >= 2 and len(list(ind2.primitives)) >= 2:
        crossover_choices.append(crossover_primitives)
    return crossover_choices
